#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Service registry and event dispatcher for Pelix.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import bisect
import logging
import threading

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Dict, List, Optional, Set, Tuple, Union
except ImportError:
    pass

# Pelix beans
from pelix.constants import (
    OBJECTCLASS,
    SERVICE_ID,
    SERVICE_RANKING,
    SERVICE_BUNDLEID,
    SERVICE_SCOPE,
    SCOPE_SINGLETON,
    SCOPE_BUNDLE,
    SCOPE_PROTOTYPE,
    BundleException,
)
from pelix.services import SERVICE_EVENT_LISTENER_HOOK
from pelix.internals.events import ServiceEvent

# Pelix utility modules
from pelix.utilities import is_string
import pelix.ldapfilter as ldapfilter

# Event hooks
from pelix.internals.hooks import ListenerInfo, ShrinkableList, ShrinkableMap

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _UsageCounter(object):
    """
    Simple reference usage counter
    """

    __slots__ = ("__count",)

    def __init__(self):
        self.__count = 0

    def inc(self):
        """
        Counter is incremented
        """
        self.__count += 1

    def dec(self):
        """
        Counter is decremented

        :return: True if the counter is still greater than 0
        """
        self.__count -= 1
        return self.__count > 0

    def is_used(self):
        """
        Tests if the reference is still used

        :return: True if the counter is still greater than 0
        """
        return self.__count > 0


# ------------------------------------------------------------------------------


class _FactoryCounter(object):
    """
    A service factory usage counter per bundle and reference
    """

    __slots__ = ("__bundle", "__factored")

    def __init__(self, bundle):
        """
        Sets up members

        :param bundle: The bundle monitored by this counter
        """
        self.__bundle = bundle

        # Service Factory Reference -> (Service instance, Usage counter)
        self.__factored = {}

    def is_used(self):
        # type: () -> bool
        """
        Checks if this counter has at least one value

        :return: True if a service is still referenced by this service
        """
        return bool(self.__factored)

    def _get_from_factory(self, factory, svc_registration):
        # type: (Any, ServiceRegistration) -> Any
        """
        Returns a service instance from a Prototype Service Factory

        :param factory: The prototype service factory
        :param svc_registration: The ServiceRegistration object
        :return: The requested service instance returned by the factory
        """
        svc_ref = svc_registration.get_reference()
        try:
            # Use the existing service
            service, counter = self.__factored[svc_ref]
            counter.inc()
        except KeyError:
            # Create the service
            service = factory.get_service(self.__bundle, svc_registration)
            counter = _UsageCounter()
            counter.inc()

            # Store the counter
            self.__factored[svc_ref] = (service, counter)

        return service

    def _get_from_prototype(self, factory, svc_registration):
        # type: (Any, ServiceRegistration) -> Any
        """
        Returns a service instance from a Prototype Service Factory

        :param factory: The service factory
        :param svc_registration: The ServiceRegistration object
        :return: The requested service instance returned by the factory
        """
        svc_ref = svc_registration.get_reference()
        service = factory.get_service(self.__bundle, svc_registration)

        try:
            # Check if the service already exists
            services, counter = self.__factored[svc_ref]
            services.append(service)
            counter.inc()
        except KeyError:
            counter = _UsageCounter()
            counter.inc()

            # Store the counter
            self.__factored[svc_ref] = ([service], counter)

        return service

    def get_service(self, factory, svc_registration):
        # type: (Any, ServiceRegistration) -> Any
        """
        Returns the service required by the bundle. The Service Factory is
        called only when necessary while the Prototype Service Factory is
        called each time

        :param factory: The service factory
        :param svc_registration: The ServiceRegistration object
        :return: The requested service instance (created if necessary)
        """
        svc_ref = svc_registration.get_reference()
        if svc_ref.is_prototype():
            return self._get_from_prototype(factory, svc_registration)

        return self._get_from_factory(factory, svc_registration)

    def unget_service(self, factory, svc_registration, service=None):
        # type: (Any, ServiceRegistration, Any) -> bool
        """
        Releases references to the given service reference

        :param factory: The service factory
        :param svc_registration: The ServiceRegistration object
        :param service: Service instance (for prototype factories)
        :return: True if all service references to this service factory
                 have been released
        """
        svc_ref = svc_registration.get_reference()
        try:
            _, counter = self.__factored[svc_ref]
        except KeyError:
            logging.warning(
                "Trying to release an unknown service factory: %s", svc_ref
            )
        else:
            if svc_ref.is_prototype():
                # Notify the factory to clean up this instance
                factory.unget_service_instance(
                    self.__bundle, svc_registration, service
                )

            if not counter.dec():
                # All references have been released: clean up
                del self.__factored[svc_ref]

                # Call the factory
                factory.unget_service(self.__bundle, svc_registration)

                # No more reference to this service
                return True

        # Some references are still there
        return False

    def cleanup_service(self, factory, svc_registration):
        # type: (Any, ServiceRegistration) -> bool
        """
        If this bundle used that factory, releases the reference; else does
        nothing

        :param factory: The service factory
        :param svc_registration: The ServiceRegistration object
        :return: True if the bundle was using the factory, else False
        """
        svc_ref = svc_registration.get_reference()
        try:
            # "service" for factories, "services" for prototypes
            services, _ = self.__factored.pop(svc_ref)
        except KeyError:
            return False
        else:
            if svc_ref.is_prototype() and services:
                for service in services:
                    try:
                        factory.unget_service_instance(
                            self.__bundle, svc_registration, service
                        )
                    except Exception:
                        # Ignore instance-level exceptions, potential errors
                        # will reappear in unget_service()
                        pass

            # Call the factory
            factory.unget_service(self.__bundle, svc_registration)

            # No more association
            svc_ref.unused_by(self.__bundle)
            return True


# ------------------------------------------------------------------------------


class ServiceReference(object):
    """
    Represents a reference to a service
    """

    __slots__ = (
        "__bundle",
        "__properties",
        "__service_id",
        "__sort_key",
        "__using_bundles",
        "_props_lock",
        "__usage_lock",
    )

    def __init__(self, bundle, properties):
        """
        :param bundle: The bundle registering the service
        :param properties: The service properties
        :raise BundleException: The properties doesn't contain mandatory
                                entries
        """
        # Check properties
        for mandatory in SERVICE_ID, OBJECTCLASS:
            if mandatory not in properties:
                raise BundleException(
                    "A Service must at least have a '{0}' entry".format(
                        mandatory
                    )
                )

        # Properties lock (used by ServiceRegistration too)
        self._props_lock = threading.RLock()

        # Usage lock
        self.__usage_lock = threading.Lock()

        # Service details
        self.__bundle = bundle
        self.__properties = properties
        self.__service_id = properties[SERVICE_ID]

        # Bundle object -> Usage Counter object
        self.__using_bundles = {}

        # Compute the sort key
        self.__sort_key = None
        self.update_sort_key()

    def __str__(self):
        """
        String representation
        """
        return "ServiceReference(ID={0}, Bundle={1}, Specs={2})".format(
            self.__service_id,
            self.__bundle.get_bundle_id(),
            self.__properties[OBJECTCLASS],
        )

    def __hash__(self):
        """
        Returns the service hash, i.e. its ID, unique in a framework instance.

        :return: The service ID
        """
        return self.__service_id

    def __eq__(self, other):
        """
        Two references are equal if they have the same service ID
        """
        # pylint: disable=W0212
        return self.__service_id == other.__service_id

    def __lt__(self, other):
        """
        Lesser than other
        """
        # pylint: disable=W0212
        return self.__sort_key < other.__sort_key

    def __gt__(self, other):
        """
        Greater than other
        """
        # pylint: disable=W0212
        return self.__sort_key > other.__sort_key

    def __le__(self, other):
        """
        Lesser than or equal to other"
        """
        # pylint: disable=W0212
        return self.__sort_key <= other.__sort_key

    def __ge__(self, other):
        """
        Greater than or equal to other
        """
        # pylint: disable=W0212
        return self.__sort_key >= other.__sort_key

    def __ne__(self, other):
        """
        Two references are different if they have different service IDs
        """
        # pylint: disable=W0212
        return self.__service_id != other.__service_id

    def get_bundle(self):
        """
        Returns the bundle that registered this service

        :return: the bundle that registered this service
        """
        return self.__bundle

    def get_using_bundles(self):
        """
        Returns the list of bundles that use this service

        :return: A list of Bundle objects
        """
        return list(self.__using_bundles.keys())

    def get_properties(self):
        """
        Returns a copy of the service properties

        :return: A copy of the service properties
        """
        with self._props_lock:
            return self.__properties.copy()

    def get_property(self, name):
        """
        Retrieves the property value for the given name

        :return: The property value, None if not found
        """
        with self._props_lock:
            return self.__properties.get(name)

    def get_property_keys(self):
        """
        Returns an array of the keys in the properties of the service

        :return: An array of property keys.
        """
        with self._props_lock:
            return tuple(self.__properties.keys())

    def is_factory(self):
        """
        Returns True if this reference points to a service factory

        :return: True if the service provides from a factory
        """
        return self.__properties[SERVICE_SCOPE] in (
            SCOPE_BUNDLE,
            SCOPE_PROTOTYPE,
        )

    def is_prototype(self):
        """
        Returns True if this reference points to a prototype service factory

        :return: True if the service provides from a prototype factory
        """
        return self.__properties[SERVICE_SCOPE] == SCOPE_PROTOTYPE

    def unused_by(self, bundle):
        """
        Indicates that this reference is not being used anymore by the given
        bundle.
        This method should only be used by the framework.

        :param bundle: A bundle that used this reference
        """
        if bundle is None or bundle is self.__bundle:
            # Ignore
            return

        with self.__usage_lock:
            try:
                if not self.__using_bundles[bundle].dec():
                    # This bundle has cleaner all of its usages of this
                    # reference
                    del self.__using_bundles[bundle]
            except KeyError:
                # Ignore error
                pass

    def used_by(self, bundle):
        """
        Indicates that this reference is being used by the given bundle.
        This method should only be used by the framework.

        :param bundle: A bundle using this reference
        """
        if bundle is None or bundle is self.__bundle:
            # Ignore
            return

        with self.__usage_lock:
            self.__using_bundles.setdefault(bundle, _UsageCounter()).inc()

    def __compute_key(self):
        """
        Computes the sort key according to the service properties

        :return: The sort key to use for this reference
        """
        return (
            -int(self.__properties.get(SERVICE_RANKING, 0)),
            self.__service_id,
        )

    def needs_sort_update(self):
        """
        Checks if the sort key must be updated

        :return: True if the sort key must be updated
        """
        return self.__sort_key != self.__compute_key()

    def update_sort_key(self):
        """
        Recomputes the sort key, based on the service ranking and ID

        See: http://www.osgi.org/javadoc/r4v43/org/osgi/framework/
                          ServiceReference.html#compareTo%28java.lang.Object%29
        """
        self.__sort_key = self.__compute_key()


# ------------------------------------------------------------------------------


class ServiceRegistration(object):
    """
    Represents a service registration object
    """

    __slots__ = (
        "__framework",
        "__reference",
        "__properties",
        "__update_callback",
    )

    def __init__(self, framework, reference, properties, update_callback):
        """
        :param framework: The host framework
        :param reference: A service reference
        :param properties: A reference to the ServiceReference properties
                           dictionary object
        :param update_callback: Method to call when the sort key is modified
        """
        self.__framework = framework
        self.__reference = reference  # type: ServiceReference
        self.__properties = properties
        self.__update_callback = update_callback

    def __str__(self):
        """
        String representation
        """
        return "ServiceRegistration({0})".format(self.__reference)

    def get_reference(self):
        # type: () -> ServiceReference
        """
        Returns the reference associated to this registration

        :return: A ServiceReference object
        """
        return self.__reference

    def set_properties(self, properties):
        """
        Updates the service properties

        :param properties: The new properties
        :raise TypeError: The argument is not a dictionary
        """
        if not isinstance(properties, dict):
            raise TypeError("Waiting for dictionary")

        # Keys that must not be updated
        for forbidden_key in OBJECTCLASS, SERVICE_ID:
            try:
                del properties[forbidden_key]
            except KeyError:
                pass

        to_delete = []
        for key, value in properties.items():
            if self.__properties.get(key) == value:
                # No update
                to_delete.append(key)

        for key in to_delete:
            # Remove unchanged properties
            del properties[key]

        if not properties:
            # Nothing to do
            return

        # Ensure that the service has a valid service ranking
        try:
            properties[SERVICE_RANKING] = int(properties[SERVICE_RANKING])
        except (ValueError, TypeError):
            # Bad value: ignore update
            del properties[SERVICE_RANKING]
        except KeyError:
            # Service ranking not updated: ignore
            pass

        # pylint: disable=W0212
        with self.__reference._props_lock:
            # Update the properties
            previous = self.__properties.copy()
            self.__properties.update(properties)

            if self.__reference.needs_sort_update():
                # The sort key and the registry must be updated
                self.__update_callback(self.__reference)

            # Trigger a new computation in the framework
            event = ServiceEvent(
                ServiceEvent.MODIFIED, self.__reference, previous
            )

            self.__framework._dispatcher.fire_service_event(event)

    def unregister(self):
        """
        Unregisters the service
        """
        self.__framework.unregister_service(self)


# ------------------------------------------------------------------------------


class EventDispatcher(object):
    """
    Simple event dispatcher
    """

    def __init__(self, registry, logger=None):
        """
        Sets up the dispatcher

        :param registry:  The service registry
        :param logger: The logger to be used
        """
        self._registry = registry

        # Logger
        self._logger = logger or logging.getLogger("EventDispatcher")

        # Bundle listeners
        self.__bnd_listeners = []
        self.__bnd_lock = threading.Lock()

        # Service listeners (specification -> listener bean)
        self.__svc_listeners = {}
        # listener instance -> listener bean
        self.__listeners_data = {}
        self.__svc_lock = threading.Lock()

        # Framework stop listeners
        self.__fw_listeners = []
        self.__fw_lock = threading.Lock()

    def clear(self):
        """
        Clears the event dispatcher
        """
        with self.__bnd_lock:
            self.__bnd_listeners = []

        with self.__svc_lock:
            self.__svc_listeners.clear()

        with self.__fw_lock:
            self.__fw_listeners = []

    def add_bundle_listener(self, listener):
        """
        Adds a bundle listener

        :param listener: The bundle listener to register
        :return: True if the listener has been registered, False if it was
                 already known
        :raise BundleException: An invalid listener has been given
        """
        if listener is None or not hasattr(listener, "bundle_changed"):
            raise BundleException("Invalid bundle listener given")

        with self.__bnd_lock:
            if listener in self.__bnd_listeners:
                self._logger.warning(
                    "Already known bundle listener '%s'", listener
                )
                return False

            self.__bnd_listeners.append(listener)
            return True

    def add_framework_listener(self, listener):
        """
        Registers a listener that will be called back right before the
        framework stops.

        :param listener: The framework stop listener
        :return: True if the listener has been registered, False if it was
                 already known
        :raise BundleException: An invalid listener has been given
        """
        if listener is None or not hasattr(listener, "framework_stopping"):
            raise BundleException("Invalid framework listener given")

        with self.__fw_lock:
            if listener in self.__fw_listeners:
                self._logger.warning(
                    "Already known framework listener '%s'", listener
                )
                return False

            self.__fw_listeners.append(listener)
            return True

    def add_service_listener(
        self, bundle_context, listener, specification=None, ldap_filter=None
    ):
        """
        Registers a service listener

        :param bundle_context: The bundle_context of the service listener
        :param listener: The service listener
        :param specification: The specification that must provide the service
                              (optional, None to accept all services)
        :param ldap_filter: Filter that must match the service properties
                            (optional, None to accept all services)
        :return: True if the listener has been registered, False if it was
                 already known
        :raise BundleException: An invalid listener has been given
        """
        if listener is None or not hasattr(listener, "service_changed"):
            raise BundleException("Invalid service listener given")

        with self.__svc_lock:
            if listener in self.__listeners_data:
                self._logger.warning(
                    "Already known service listener '%s'", listener
                )
                return False

            try:
                ldap_filter = ldapfilter.get_ldap_filter(ldap_filter)
            except ValueError as ex:
                raise BundleException("Invalid service filter: {0}".format(ex))

            stored = ListenerInfo(
                bundle_context, listener, specification, ldap_filter
            )
            self.__listeners_data[listener] = stored
            self.__svc_listeners.setdefault(specification, []).append(stored)
            return True

    def remove_bundle_listener(self, listener):
        """
        Unregisters a bundle listener

        :param listener: The bundle listener to unregister
        :return: True if the listener has been unregistered, else False
        """
        with self.__bnd_lock:
            if listener not in self.__bnd_listeners:
                return False

            self.__bnd_listeners.remove(listener)
            return True

    def remove_framework_listener(self, listener):
        """
        Unregisters a framework stop listener

        :param listener: The framework listener to unregister
        :return: True if the listener has been unregistered, else False
        """
        with self.__fw_lock:
            try:
                self.__fw_listeners.remove(listener)
                return True
            except ValueError:
                return False

    def remove_service_listener(self, listener):
        """
        Unregisters a service listener

        :param listener: The service listener
        :return: True if the listener has been unregistered
        """
        with self.__svc_lock:
            try:
                data = self.__listeners_data.pop(listener)
                spec_listeners = self.__svc_listeners[data.specification]
                spec_listeners.remove(data)
                if not spec_listeners:
                    del self.__svc_listeners[data.specification]
                return True
            except KeyError:
                return False

    def fire_bundle_event(self, event):
        """
        Notifies bundle events listeners of a new event in the calling thread.

        :param event: The bundle event
        """
        with self.__bnd_lock:
            # Copy the list of listeners
            listeners = self.__bnd_listeners[:]

        # Call'em all
        for listener in listeners:
            try:
                listener.bundle_changed(event)
            except:
                self._logger.exception("Error calling a bundle listener")

    def fire_framework_stopping(self):
        """
        Calls all framework listeners, telling them that the framework is
        stopping
        """
        with self.__fw_lock:
            # Copy the list of listeners
            listeners = self.__fw_listeners[:]

        for listener in listeners:
            try:
                listener.framework_stopping()
            except:
                self._logger.exception(
                    "An error occurred calling one of the "
                    "framework stop listeners"
                )

    def fire_service_event(self, event):
        """
        Notifies service events listeners of a new event in the calling thread.

        :param event: The service event
        """
        # Get the service properties
        properties = event.get_service_reference().get_properties()
        svc_specs = properties[OBJECTCLASS]
        previous = None
        endmatch_event = None
        svc_modified = event.get_kind() == ServiceEvent.MODIFIED

        if svc_modified:
            # Modified service event : prepare the end match event
            previous = event.get_previous_properties()
            endmatch_event = ServiceEvent(
                ServiceEvent.MODIFIED_ENDMATCH,
                event.get_service_reference(),
                previous,
            )

        with self.__svc_lock:
            # Get the listeners for this specification
            listeners = set()
            for spec in svc_specs:
                try:
                    listeners.update(self.__svc_listeners[spec])
                except KeyError:
                    pass

            # Add those which listen to any specification
            try:
                listeners.update(self.__svc_listeners[None])
            except KeyError:
                pass

        # Filter listeners with EventListenerHooks
        listeners = self._filter_with_hooks(event, listeners)

        # Get the listeners for this specification
        for data in listeners:
            # Default event to send : the one we received
            sent_event = event

            # Test if the service properties matches the filter
            ldap_filter = data.ldap_filter
            if ldap_filter is not None and not ldap_filter.matches(properties):
                # Event doesn't match listener filter...
                if (
                    svc_modified
                    and previous is not None
                    and ldap_filter.matches(previous)
                ):
                    # ... but previous properties did match
                    sent_event = endmatch_event
                else:
                    # Didn't match before either, ignore it
                    continue

            # Call'em
            try:
                data.listener.service_changed(sent_event)
            except:
                self._logger.exception("Error calling a service listener")

    def _filter_with_hooks(self, svc_event, listeners):
        """
        Filters listeners with EventListenerHooks

        :param svc_event: ServiceEvent being triggered
        :param listeners: Listeners to filter
        :return: A list of listeners with hook references
        """
        svc_ref = svc_event.get_service_reference()
        # Get EventListenerHooks service refs from registry
        hook_refs = self._registry.find_service_references(
            SERVICE_EVENT_LISTENER_HOOK
        )
        # only do something if there are some hook_refs
        if hook_refs:
            # Associate bundle context to hooks
            ctx_listeners = {}
            for listener in listeners:
                context = listener.bundle_context
                ctx_listeners.setdefault(context, []).append(listener)

            # Convert the dictionary to a shrinkable one,
            # with shrinkable lists of listeners
            shrinkable_ctx_listeners = ShrinkableMap(
                {
                    context: ShrinkableList(value)
                    for context, value in ctx_listeners.items()
                }
            )

            for hook_ref in hook_refs:
                if not svc_ref == hook_ref:
                    # Get the bundle of the hook service
                    hook_bundle = hook_ref.get_bundle()
                    # lookup service from registry
                    hook_svc = self._registry.get_service(hook_bundle, hook_ref)
                    if hook_svc is not None:
                        # call event method of the hook service,
                        # pass in svc_event and shrinkable_ctx_listeners
                        # (which can be modified by hook)
                        try:
                            hook_svc.event(svc_event, shrinkable_ctx_listeners)
                        except:
                            self._logger.exception(
                                "Error calling EventListenerHook"
                            )
                        finally:
                            # Clean up the service
                            self._registry.unget_service(hook_bundle, hook_ref)

            # Convert the shrinkable_ctx_listeners back to a list of listeners
            # before returning
            ret_listeners = set()
            for bnd_listeners in shrinkable_ctx_listeners.values():
                ret_listeners.update(bnd_listeners)

            return ret_listeners

        # No hook ref
        return listeners


# ------------------------------------------------------------------------------


class ServiceRegistry(object):
    """
    Service registry for Pelix.

    Associates service references to instances and bundles.
    """

    def __init__(self, framework, logger=None):
        """
        Sets up the registry

        :param framework: Associated framework
        :param logger: Logger to use
        """
        # Associated framework
        self.__framework = framework

        # Logger
        self._logger = logger or logging.getLogger("ServiceRegistry")

        # Next service ID
        self.__next_service_id = 1

        # Service reference -> Service instance
        self.__svc_registry = {}  # type: Dict[ServiceReference, Any]

        # Service reference -> (Service factory, Service Registration)
        self.__svc_factories = {}  # type: Dict[Any, Tuple[Any, ServiceRegistration]]

        # Specification -> Service references[] (always sorted)
        self.__svc_specs = {}

        # Services published: Bundle -> set(Service references)
        self.__bundle_svc = {}  # type: Dict[Any, Set[ServiceReference]]

        # Services consumed: Bundle -> {Service reference -> UsageCounter}
        self.__bundle_imports = {}  # type: Dict[Any, Dict[ServiceReference, _UsageCounter]]

        # Service factories consumption: Bundle -> _FactoryCounter
        self.__factory_usage = {}  # type: Dict[Any, _FactoryCounter]

        # Locks
        self.__svc_lock = threading.RLock()

        # Pending unregistration: Service reference -> Service instance
        self.__pending_services = {}  # type: Dict[ServiceReference, Any]

    def clear(self):
        """
        Clears the registry
        """
        with self.__svc_lock:
            self.__svc_registry.clear()
            self.__svc_factories.clear()
            self.__svc_specs.clear()
            self.__bundle_svc.clear()
            self.__bundle_imports.clear()
            self.__factory_usage.clear()
            self.__pending_services.clear()

    def register(
        self, bundle, classes, properties, svc_instance, factory, prototype
    ):
        """
        Registers a service.

        :param bundle: The bundle that registers the service
        :param classes: The classes implemented by the service
        :param properties: The properties associated to the service
        :param svc_instance: The instance of the service
        :param factory: If True, the given service is a service factory
        :param prototype: If True, the given service is a prototype service
                          factory (the factory argument is considered True)
        :return: The ServiceRegistration object
        """
        with self.__svc_lock:
            # Prepare properties
            service_id = self.__next_service_id
            self.__next_service_id += 1
            properties[OBJECTCLASS] = classes
            properties[SERVICE_ID] = service_id
            properties[SERVICE_BUNDLEID] = bundle.get_bundle_id()

            # Compute service scope
            if prototype:
                properties[SERVICE_SCOPE] = SCOPE_PROTOTYPE
            elif factory:
                properties[SERVICE_SCOPE] = SCOPE_BUNDLE
            else:
                properties[SERVICE_SCOPE] = SCOPE_SINGLETON

            # Force to have a valid service ranking
            try:
                properties[SERVICE_RANKING] = int(properties[SERVICE_RANKING])
            except (KeyError, ValueError, TypeError):
                properties[SERVICE_RANKING] = 0

            # Make the service reference
            svc_ref = ServiceReference(bundle, properties)

            # Make the service registration
            svc_registration = ServiceRegistration(
                self.__framework, svc_ref, properties, self.__sort_registry
            )

            # Store service information
            if prototype or factory:
                self.__svc_factories[svc_ref] = (svc_instance, svc_registration)

            # Also store factories, as they must appear like any other service
            self.__svc_registry[svc_ref] = svc_instance

            for spec in classes:
                spec_refs = self.__svc_specs.setdefault(spec, [])
                bisect.insort_left(spec_refs, svc_ref)

            # Reverse map, to ease bundle/service association
            bundle_services = self.__bundle_svc.setdefault(bundle, set())
            bundle_services.add(svc_ref)
            return svc_registration

    def __sort_registry(self, svc_ref):
        # type: (ServiceReference) -> None
        """
        Sorts the registry, after the update of the sort key of given service
        reference

        :param svc_ref: A service reference with a modified sort key
        """
        with self.__svc_lock:
            if svc_ref not in self.__svc_registry:
                raise BundleException("Unknown service: {0}".format(svc_ref))

            # Remove current references
            for spec in svc_ref.get_property(OBJECTCLASS):
                # Use bisect to remove the reference (faster)
                spec_refs = self.__svc_specs[spec]
                idx = bisect.bisect_left(spec_refs, svc_ref)
                del spec_refs[idx]

            # ... use the new sort key
            svc_ref.update_sort_key()

            for spec in svc_ref.get_property(OBJECTCLASS):
                # ... and insert it again
                spec_refs = self.__svc_specs[spec]
                bisect.insort_left(spec_refs, svc_ref)

    def unregister(self, svc_ref):
        # type: (ServiceReference) -> Any
        """
        Unregisters a service

        :param svc_ref: A service reference
        :return: The unregistered service instance
        :raise BundleException: Unknown service reference
        """
        with self.__svc_lock:
            try:
                # Try in pending services
                return self.__pending_services.pop(svc_ref)
            except KeyError:
                # Not pending: continue
                pass

            if svc_ref not in self.__svc_registry:
                raise BundleException("Unknown service: {0}".format(svc_ref))

            # Get the owner
            bundle = svc_ref.get_bundle()

            # Get the service instance
            service = self.__svc_registry.pop(svc_ref)

            for spec in svc_ref.get_property(OBJECTCLASS):
                spec_services = self.__svc_specs[spec]
                # Use bisect to remove the reference (faster)
                idx = bisect.bisect_left(spec_services, svc_ref)
                del spec_services[idx]
                if not spec_services:
                    del self.__svc_specs[spec]

            # Remove the service factory
            if svc_ref.is_factory():
                # Call unget_service for all client bundle
                factory, svc_reg = self.__svc_factories.pop(svc_ref)
                for counter in self.__factory_usage.values():
                    counter.cleanup_service(factory, svc_reg)
            else:
                # Delete bundle association
                bundle_services = self.__bundle_svc[bundle]
                bundle_services.remove(svc_ref)
                if not bundle_services:
                    # Don't keep empty lists
                    del self.__bundle_svc[bundle]

            return service

    def hide_bundle_services(self, bundle):
        """
        Hides the services of the given bundle (removes them from lists, but
        lets them be unregistered)

        :param bundle: The bundle providing services
        :return: The references of the hidden services
        """
        with self.__svc_lock:
            try:
                svc_refs = self.__bundle_svc.pop(bundle)
            except KeyError:
                # Nothing to do
                return set()
            else:
                # Clean the registry
                specs = set()
                for svc_ref in svc_refs:
                    if svc_ref.is_factory():
                        continue

                    # Remove direct references
                    self.__pending_services[svc_ref] = self.__svc_registry.pop(
                        svc_ref
                    )
                    specs.update(svc_ref.get_property(OBJECTCLASS))

                    # Clean the specifications cache
                    for spec in svc_ref.get_property(OBJECTCLASS):
                        spec_services = self.__svc_specs[spec]
                        # Use bisect to remove the reference (faster)
                        idx = bisect.bisect_left(spec_services, svc_ref)
                        del spec_services[idx]
                        if not spec_services:
                            del self.__svc_specs[spec]

            return svc_refs

    def find_service_references(
        self, clazz=None, ldap_filter=None, only_one=False
    ):
        """
        Finds all services references matching the given filter.

        :param clazz: Class implemented by the service
        :param ldap_filter: Service filter
        :param only_one: Return the first matching service reference only
        :return: A list of found references, or None
        :raise BundleException: An error occurred looking for service
                                references
        """
        with self.__svc_lock:
            if clazz is None and ldap_filter is None:
                # Return a sorted copy of the keys list
                # Do not return None, as the whole content was required
                return sorted(self.__svc_registry.keys())

            if hasattr(clazz, "__name__"):
                # Escape the type name
                clazz = ldapfilter.escape_LDAP(clazz.__name__)
            elif is_string(clazz):
                # Escape the class name
                clazz = ldapfilter.escape_LDAP(clazz)

            if clazz is None:
                # Directly use the given filter
                refs_set = sorted(self.__svc_registry.keys())
            else:
                try:
                    # Only for references with the given specification
                    refs_set = iter(self.__svc_specs[clazz])
                except KeyError:
                    # No matching specification
                    return None

            # Parse the filter
            try:
                new_filter = ldapfilter.get_ldap_filter(ldap_filter)
            except ValueError as ex:
                raise BundleException(ex)

            if new_filter is not None:
                # Prepare a generator, as we might not need a complete
                # walk-through
                refs_set = (
                    ref
                    for ref in refs_set
                    if new_filter.matches(ref.get_properties())
                )

            if only_one:
                # Return the first element in the list/generator
                try:
                    return [next(refs_set)]
                except StopIteration:
                    # No match
                    return None

            # Get all the matching references
            return list(refs_set) or None

    def get_bundle_imported_services(self, bundle):
        """
        Returns this bundle's ServiceReference list for all services it is
        using or returns None if this bundle is not using any services.
        A bundle is considered to be using a service if its use count for that
        service is greater than zero.

        The list is valid at the time of the call to this method, however, as
        the Framework is a very dynamic environment, services can be modified
        or unregistered at any time.

        :param bundle: The bundle to look into
        :return: The references of the services used by this bundle
        """
        with self.__svc_lock:
            return sorted(self.__bundle_imports.get(bundle, []))

    def get_bundle_registered_services(self, bundle):
        # type: (Any) -> List[ServiceReference]
        """
        Retrieves the services registered by the given bundle. Returns None
        if the bundle didn't register any service.

        :param bundle: The bundle to look into
        :return: The references to the services registered by the bundle
        """
        with self.__svc_lock:
            return sorted(self.__bundle_svc.get(bundle, []))

    def get_service(self, bundle, reference):
        # type: (Any, ServiceReference) -> Any
        """
        Retrieves the service corresponding to the given reference

        :param bundle: The bundle requiring the service
        :param reference: A service reference
        :return: The requested service
        :raise BundleException: The service could not be found
        """
        with self.__svc_lock:
            if reference.is_factory():
                return self.__get_service_from_factory(bundle, reference)

            # Be sure to have the instance
            try:
                service = self.__svc_registry[reference]

                # Indicate the dependency
                imports = self.__bundle_imports.setdefault(bundle, {})
                imports.setdefault(reference, _UsageCounter()).inc()
                reference.used_by(bundle)
                return service
            except KeyError:
                # Not found
                raise BundleException(
                    "Service not found (reference: {0})".format(reference)
                )

    def __get_service_from_factory(self, bundle, reference):
        # type: (Any, ServiceReference) -> Any
        """
        Returns a service instance from a service factory or a prototype
        service factory

        :param bundle: The bundle requiring the service
        :param reference: A reference pointing to a factory
        :return: The requested service
        :raise BundleException: The service could not be found
        """
        try:
            factory, svc_reg = self.__svc_factories[reference]

            # Indicate the dependency
            imports = self.__bundle_imports.setdefault(bundle, {})
            if reference not in imports:
                # New reference usage: store a single usage
                # The Factory counter will handle the rest
                usage_counter = _UsageCounter()
                usage_counter.inc()
                imports[reference] = usage_counter
                reference.used_by(bundle)

            # Check the per-bundle usage counter
            factory_counter = self.__factory_usage.setdefault(
                bundle, _FactoryCounter(bundle)
            )
            return factory_counter.get_service(factory, svc_reg)
        except KeyError:
            # Not found
            raise BundleException(
                "Service not found (reference: {0})".format(reference)
            )

    def unget_used_services(self, bundle):
        """
        Cleans up all service usages of the given bundle.

        :param bundle: Bundle to be cleaned up
        """
        # Pop used references
        try:
            imported_refs = list(self.__bundle_imports.pop(bundle))
        except KeyError:
            # Nothing to do
            return

        for svc_ref in imported_refs:
            # Remove usage marker
            svc_ref.unused_by(bundle)

            if svc_ref.is_prototype():
                # Get factory information and clean up the service from the
                # factory counter
                factory_counter = self.__factory_usage.pop(bundle)
                factory, svc_reg = self.__svc_factories[svc_ref]
                factory_counter.cleanup_service(factory, svc_reg)
            elif svc_ref.is_factory():
                # Factory service, release it the standard way
                self.__unget_service_from_factory(bundle, svc_ref)

        # Clean up local structures
        try:
            del self.__factory_usage[bundle]
        except KeyError:
            pass

        try:
            self.__bundle_imports.pop(bundle).clear()
        except KeyError:
            pass

    def unget_service(self, bundle, reference, service=None):
        # type: (Any, ServiceReference, Any) -> bool
        """
        Removes the usage of a service by a bundle

        :param bundle: The bundle that used the service
        :param reference: A service reference
        :param service: Service instance (for Prototype Service Factories)
        :return: True if the bundle usage has been removed
        """
        with self.__svc_lock:
            if reference.is_prototype():
                return self.__unget_service_from_factory(
                    bundle, reference, service
                )
            elif reference.is_factory():
                return self.__unget_service_from_factory(bundle, reference)

            try:
                # Remove the service reference from the bundle
                imports = self.__bundle_imports[bundle]
                if not imports[reference].dec():
                    # No more reference to it
                    del imports[reference]
            except KeyError:
                # Unknown reference
                return False
            else:
                # Clean up
                if not imports:
                    del self.__bundle_imports[bundle]

                # Update the service reference
                reference.unused_by(bundle)
                return True

    def __unget_service_from_factory(self, bundle, reference, service=None):
        # type: (Any, ServiceReference, Any) -> bool
        """
        Removes the usage of a a service factory or a prototype
        service factory by a bundle

        :param bundle: The bundle that used the service
        :param reference: A service reference
        :param service: Service instance (for prototype factories)
        :return: True if the bundle usage has been removed
        """
        try:
            factory, svc_reg = self.__svc_factories[reference]
        except KeyError:
            # Unknown service reference
            return False

        # Check the per-bundle usage counter
        try:
            counter = self.__factory_usage[bundle]
        except KeyError:
            # Unknown reference to a factory
            return False
        else:
            if counter.unget_service(factory, svc_reg, service):
                try:
                    # No more dependency
                    reference.unused_by(bundle)

                    # All references have been taken away: clean up
                    if not self.__factory_usage[bundle].is_used():
                        del self.__factory_usage[bundle]

                    # Remove the service reference from the bundle
                    imports = self.__bundle_imports[bundle]
                    del imports[reference]
                except KeyError:
                    # Unknown reference
                    return False
                else:
                    # Clean up
                    if not imports:
                        del self.__bundle_imports[bundle]

        return True



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix remote services: ZooKeeper-based discovery and event notification

*Note:* This discovery package requires the ``kazoo`` package (available on
PyPI), and a ZooKeeper server.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import posixpath
import socket
import threading
import uuid

# ZooKeeper library
from kazoo.client import KazooClient, KazooState, EventType
from kazoo.client import WatchedEvent  # pylint:disable=unused-import
from kazoo.exceptions import KazooException, NodeExistsError

# iPOPO decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    Invalidate,
    Validate,
    Property,
)

# Pelix utilities
import pelix.constants
from pelix.utilities import to_bytes, to_str
from pelix.threadpool import ThreadPool

# Remote services
from pelix.remote.edef_io import EDEFWriter, EDEFReader
import pelix.remote
import pelix.remote.beans as beans

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

_logger = logging.getLogger(__name__)
""" Module's logger """

FRAMEWORKS_ROOT = "/frameworks"
""" Path to the Z-node parent of frameworks definition Z-node """

ENDPOINTS_ROOT = "/endpoints"
""" Path to the Z-node containing the endpoints of each framework """

# ------------------------------------------------------------------------------


class ZooKeeperClient:
    """
    Handles basic ZooKeeper events
    """

    def __init__(self, zk_hosts, log_name="ZooKeeperClient", prefix=""):
        """
        :param zk_hosts: List of ZooKeepers hosts
        :param log_name: Name of the logger to use
        :param prefix: Prefix to all paths
        """
        self._zk = KazooClient(zk_hosts)
        self._zk.add_listener(self.__conn_listener)
        self.__prefix = prefix

        # Session state
        self.__connected = False
        self.__online = False
        self.__stop = False

        # Notification queue
        self._logger = logging.getLogger(log_name)
        self._queue = ThreadPool(1, 1, logname=log_name)

    @property
    def prefix(self):
        """
        Prefix to all ZooKeeper nodes
        """
        return self.__prefix

    @property
    def connected(self):
        """
        ZooKeeper client state: connected to the quorum
        """
        return self.__connected

    @property
    def online(self):
        """
        ZooKeeper client state: connected & online (session active)
        """
        return self.__online

    @property
    def stopped(self):
        """
        ZooKeeper client status (stop requested)
        """
        return self.__stop

    def __conn_listener(self, state):
        """
        Connection event listener

        :param state: The new connection state
        """
        if state == KazooState.CONNECTED:
            self.__online = True
            if not self.__connected:
                self.__connected = True
                self._logger.info("Connected to ZooKeeper")
                self._queue.enqueue(self.on_first_connection)
            else:
                self._logger.warning("Re-connected to ZooKeeper")
                self._queue.enqueue(self.on_client_reconnection)
        elif state == KazooState.SUSPENDED:
            self._logger.warning("Connection suspended")
            self.__online = False
        elif state == KazooState.LOST:
            self.__online = False
            self.__connected = False

            if self.__stop:
                self._logger.info("Disconnected from ZooKeeper (requested)")
            else:
                self._logger.warning("Connection lost")

    def start(self):
        """
        Starts the connection
        """
        self.__stop = False
        self._queue.start()
        self._zk.start()

    def stop(self):
        """
        Stops the connection
        """
        self.__stop = True
        self._queue.stop()
        self._zk.stop()

    @staticmethod
    def on_first_connection():
        """
        Called when the client is connected for the first time
        """

    @staticmethod
    def on_client_reconnection():
        """
        Called when the client is reconnected to the server
        """

    def __path(self, path):
        """
        Adds the prefix to the given path

        :param path: Z-Path
        :return: Prefixed Z-Path
        """
        if path.startswith(self.__prefix):
            return path

        return "{}{}".format(self.__prefix, path)

    def create(self, path, data, ephemeral=False, sequence=False):
        """
        Creates a ZooKeeper node

        :param path: Z-Path
        :param data: Node Content
        :param ephemeral: Ephemeral flag
        :param sequence: Sequential flag
        """
        return self._zk.create(
            self.__path(path), data, ephemeral=ephemeral, sequence=sequence
        )

    def ensure_path(self, path):
        """
        Ensures that a path exists, creates it if necessary

        :param path: Z-Path
        """
        return self._zk.ensure_path(self.__path(path))

    def get(self, path, watch=None):
        """
        Gets the content of a ZooKeeper node

        :param path: Z-Path
        :param watch: Watch method
        """
        return self._zk.get(self.__path(path), watch=watch)

    def get_children(self, path, watch=None):
        """
        Gets the list of children of a node

        :param path: Z-Path
        :param watch: Watch method
        """
        return self._zk.get_children(self.__path(path), watch=watch)

    def set(self, path, data):
        """
        Sets the content of a ZooKeeper node

        :param path: Z-Path
        :param data: New content
        """
        return self._zk.set(self.__path(path), data)

    def delete(self, path):
        """
        Deletes a node

        :param path: Z-Path
        """
        return self._zk.delete(self.__path(path))


# ------------------------------------------------------------------------------


@ComponentFactory(pelix.remote.FACTORY_DISCOVERY_ZOOKEEPER)
@Provides(
    pelix.remote.SERVICE_EXPORT_ENDPOINT_LISTENER, controller="_controller"
)
@Requires("_dispatcher", pelix.remote.SERVICE_DISPATCHER)
@Requires("_registry", pelix.remote.SERVICE_REGISTRY)
@Property("_prefix", "zookeeper.prefix", "/pelix")
@Property("_zk_hosts", "zookeeper.hosts", "localhost:2181")
class ZooKeeperDiscovery(object):
    """
    Pelix Remote Service discovery provider based on ZooKeeper
    """

    def __init__(self):
        # Requirements
        self._dispatcher = None
        self._registry = None

        # Properties
        self._controller = False
        self._prefix = None
        self._zk_hosts = None
        self._zk = None  # type: ZooKeeperClient

        # Framework properties
        self._fw_uid = None

        # Session state
        self.__uuid = str(uuid.uuid4())
        self.__master = threading.Event()

        # Keep track of frameworks hosts
        self._frameworks_hosts = {}

    @Validate
    def _validate(self, context):
        """
        Component validated
        """
        self._controller = False

        # Get the framework UID
        self._fw_uid = context.get_property(pelix.constants.FRAMEWORK_UID)

        # Check configuration
        if not self._zk_hosts:
            raise ValueError("ZooKeeper hosts configuration is missing")

        # Start ZooKeeper client
        self._zk = ZooKeeperClient(self._zk_hosts, _logger.name, self._prefix)
        self._zk.on_first_connection = self._on_first_connection
        self._zk.start()
        _logger.debug("ZooKeeper Discovery validated")

    @staticmethod
    def _endpoint_path(fw_uid, endpoint_uid):
        # type: (str, str) -> str
        """
        Returns the path to the given endpoint

        :param fw_uid: Framework UID (host of endpoint)
        :param endpoint_uid: Endpoint UID
        :return: The path to the endpoint Z-node
        """
        return posixpath.join(ENDPOINTS_ROOT, fw_uid, endpoint_uid)

    @staticmethod
    def _endpoints_path(fw_uid):
        # type: (str) -> str
        """
        Returns the path to the endpoints parent Z-node for the given framework

        :param fw_uid: Framework UID (host of endpoints)
        :return: The path to the endpoints Z-node
        """
        return posixpath.join(ENDPOINTS_ROOT, fw_uid)

    @staticmethod
    def _framework_path(fw_uid):
        # type: (str) -> str
        """
        Returns the path to the framework Z-node

        :param fw_uid: Framework UID
        :return: The path to the framework Z-node
        """
        return posixpath.join(FRAMEWORKS_ROOT, fw_uid)

    @Invalidate
    def _invalidate(self, _):
        """
        Component invalidated
        """
        self._controller = False

        # Clean up properly
        self._clear_framework(self._fw_uid)

        # Stop the Kazoo Client
        self._zk.stop()

    def _clear_framework(self, fw_uid):
        # type: (str) -> None
        """
        Clears all references to the given framework

        :param fw_uid: A framework UID
        """
        # Clear all endpoints
        fw_endpoints_path = self._endpoints_path(fw_uid)
        try:
            # Get the list of endpoints for the given framework
            endpoints = self._zk.get_children(fw_endpoints_path)
        except KazooException:
            # Already done
            pass
        else:
            for endpoint_uid in endpoints:
                try:
                    # Delete each endpoint
                    self._zk.delete(self._endpoint_path(fw_uid, endpoint_uid))
                except KazooException:
                    # Already done
                    pass

        # Clear remaining paths
        for root_path in (fw_endpoints_path, self._framework_path(fw_uid)):
            try:
                self._zk.delete(root_path)
            except KazooException:
                # Already done
                pass

    def _on_first_connection(self):
        """
        Called on first connection to ZooKeeper
        """
        _logger.debug("Connected to ZooKeeper")

        # Ensure paths
        for path in (
            FRAMEWORKS_ROOT,
            ENDPOINTS_ROOT,
            self._endpoints_path(self._fw_uid),
        ):
            self._zk.ensure_path(path)

        # Register the framework
        self._register_framework()

        # Listen to local events
        self._controller = True

    def _register_framework(self):
        """
        Registers the framework and its current services in Redis
        """
        # The framework key
        fw_node = self._framework_path(self._fw_uid)

        # The host name
        hostname = socket.gethostname()
        if hostname == "localhost":
            logging.warning(
                "Hostname is '%s': this will be a problem for "
                "multi-host remote services",
                hostname,
            )

        # Prepare an ephemeral Z-Node
        self._zk.create(fw_node, to_bytes(hostname), True)

        # Load existing services
        self._load_existing_endpoints()

        # Register all already exported services
        for endpoint in self._dispatcher.get_endpoints():
            self._register_service(endpoint)

    def _cache_fw_host(self, fw_uid):
        # type: (str) -> str
        """
        Gets the host name associated to a framework. Caches it if necessary.
        Also, adds a watcher on the framework Z-Node

        :param fw_uid: UID of a framework
        :return: The framework host name
        """
        try:
            return self._frameworks_hosts[fw_uid]
        except KeyError:
            fw_host = self._frameworks_hosts[fw_uid] = to_str(
                self._zk.get(
                    self._framework_path(fw_uid), self._on_framework_event
                )[0]
            )
            return fw_host

    def __read_endpoint(self, path):
        # type: (str) -> beans.EndpointDescription
        """
        Reads the description of an endpoint at the given Z-Node path.
        Also set the endpoint event listener on the node.

        :param path: Path to the Z-Node describing the endpoint
        :return: An EndpointDescription bean
        """
        return EDEFReader().parse(
            to_str(self._zk.get(path, self._on_endpoint_event)[0])
        )[0]

    def _load_existing_endpoints(self):
        """
        Loads already-registered endpoints
        """
        # Get the list of frameworks
        frameworks = self._zk.get_children(
            FRAMEWORKS_ROOT, watch=self._on_frameworks_event
        )

        for fw_uid in frameworks:
            # Avoid ourselves
            if fw_uid == self._fw_uid:
                continue

            # Store the host name of this framework
            self._cache_fw_host(fw_uid)

            # List all endpoints of this framework
            endpoints = self._zk.get_children(
                self._endpoints_path(fw_uid), self._on_fw_endpoints_event
            )

            for endpoint_uid in endpoints:
                # Get JSON description
                endpoint = self.__read_endpoint(
                    self._endpoint_path(fw_uid, endpoint_uid)
                )

                # Register the remote service
                self._register_remote(endpoint)

    def _register_service(self, endpoint):
        # type: (beans.ExportEndpoint) -> None
        """
        Register a local endpoint

        :param endpoint: A local endpoint
        """
        # Prepare node content
        path = self._endpoint_path(self._fw_uid, endpoint.uid)
        data = to_bytes(
            EDEFWriter().to_string(
                [beans.EndpointDescription.from_export(endpoint)]
            )
        )

        try:
            try:
                # Create an ephemeral node
                self._zk.create(path, data, True)
            except NodeExistsError:
                # Service already exists: update it
                self._zk.set(path, data)
        except KazooException as ex:
            _logger.warning(
                "Error registering local service: %s", type(ex).__name__
            )

    def _unregister_service(self, endpoint):
        # type: (beans.ExportEndpoint) -> None
        """
        Unregisters an endpoint from Redis

        :param endpoint: A :class:`~pelix.remote.ExportEndpoint` object
        """
        try:
            self._zk.delete(self._endpoint_path(self._fw_uid, endpoint.uid))
        except KazooException as ex:
            _logger.error("Error unregistering service %s:", ex)

    def endpoints_added(self, endpoints):
        # type: (list) -> None
        """
        Multiple endpoints have been added

        :param endpoints: A list of ExportEndpoint beans
        """
        for endpoint in endpoints:  # type: beans.ExportEndpoint
            self._register_service(endpoint)

    def endpoint_updated(self, endpoint, _):
        # type: (beans.ExportEndpoint, dict) -> None
        """
        An end point is updated

        :param endpoint: The updated endpoint
                         (:class:`~pelix.remote.ExportEndpoint`)
        :param _: Previous end point properties (not used)
        """
        # Update and registration are the same with Redis
        self._register_service(endpoint)

    def endpoint_removed(self, endpoint):
        # type: (beans.ExportEndpoint) -> None
        """
        An end point is removed

        :param endpoint: :class:`~pelix.remote.ExportEndpoint` being removed
        """
        self._unregister_service(endpoint)

    def _register_remote(self, endpoint_desc):
        # type: (beans.EndpointDescription) -> bool
        """
        Registers a discovered remote endpoint

        :param endpoint_desc: A remote endpoint description
        """
        # Get the host of the parent framework
        fw_uid = endpoint_desc.get_framework_uuid()
        fw_host = self._cache_fw_host(fw_uid)

        # Register the endpoint
        endpoint = endpoint_desc.to_import()
        endpoint.server = fw_host
        if self._registry.contains(endpoint):
            # Update endpoint
            self._registry.update(endpoint.uid, endpoint.properties)
        else:
            # New endpoint
            self._registry.add(endpoint)
        return True

    def _on_frameworks_event(self, event):
        # type: (WatchedEvent) -> None
        """
        Handles an event on the Frameworks Z-node
        """
        if event.type == EventType.CHILD:
            self.__check_frameworks()
        else:
            # Reset the listener in any case
            self._zk.get(event.path, self._on_frameworks_event)

    def _on_framework_event(self, event):
        # type: (WatchedEvent) -> None
        """
        Handles an event on a framework's Z-node
        """
        if event.type == EventType.CHANGED:
            _logger.warning("Unhandled change of host name for a framework")
        elif event.type == EventType.DELETED:
            # Framework is gone
            fw_uid = event.path.rsplit("/", 1)[-1]
            self._registry.lost_framework(fw_uid)
        else:
            # Reset the listener in any case
            self._zk.get(event.path, self._on_framework_event)

    def __check_frameworks(self):
        """
        Checks the list of frameworks in ZooKeeper
        """
        try:
            # Get the list of frameworks
            frameworks = self._zk.get_children(
                FRAMEWORKS_ROOT, watch=self._on_frameworks_event
            )
        except KazooException:
            # Error reading the list of frameworks
            _logger.warning("Error looking for new frameworks")
            return

        # Find delta with our current state
        to_add = set(frameworks) - set(self._frameworks_hosts) - {self._fw_uid}
        to_del = set(self._frameworks_hosts) - set(frameworks)

        # Consider frameworks lost (the registry will handle the endpoints)
        for fw_uid in to_del:
            del self._frameworks_hosts[fw_uid]
            self._registry.lost_framework(fw_uid)

        # Get the host names of the new frameworks
        for fw_uid in to_add:
            self._cache_fw_host(fw_uid)

            endpoints = self._zk.get_children(
                self._endpoints_path(fw_uid), self._on_fw_endpoints_event
            )

            for endpoint_uid in endpoints:
                endpoint = self.__read_endpoint(
                    self._endpoint_path(fw_uid, endpoint_uid)
                )
                self._register_remote(endpoint)

    def _on_fw_endpoints_event(self, event):
        # type: (WatchedEvent) -> None
        """
        Handles an event on the endpoints Z-node of a framework
        """
        if event.type == EventType.CHILD:
            fw_uid = event.path.rsplit("/", 1)[-1]
            self.__check_framework_endpoints(fw_uid)
        elif event.type != EventType.DELETED:
            # Reset the listener in any case
            self._zk.get(event.path, self._on_fw_endpoints_event)

    def _on_endpoint_event(self, event):
        # type: (WatchedEvent) -> None
        """
        Handles an event on an endpoint Z-node
        """
        if event.type == EventType.CHANGED:
            # Endpoint updated
            endpoint = self.__read_endpoint(event.path)
            self._register_remote(endpoint)
        elif event.type == EventType.DELETED:
            # Endpoint deleted
            endpoint_uid = event.path.rsplit("/", 1)[-1]
            self._registry.remove(endpoint_uid)

    def __check_framework_endpoints(self, fw_uid):
        # type: (str) -> None
        """
        Checks the list of endpoints for a framework in ZooKeeper
        """
        try:
            # Get the list of frameworks
            endpoints = self._zk.get_children(
                self._endpoints_path(fw_uid), self._on_fw_endpoints_event
            )
        except KazooException:
            # No more endpoints for this framework
            return

        for endpoint_uid in endpoints:
            if not self._registry.contains(endpoint_uid):
                try:
                    # New endpoint found
                    self._register_remote(
                        self.__read_endpoint(
                            self._endpoint_path(fw_uid, endpoint_uid)
                        )
                    )
                except KazooException:
                    logging.warning(
                        "Error reading endpoint %s of framework %s",
                        fw_uid,
                        endpoint_uid,
                    )



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Utility methods and decorators

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import collections
import contextlib
import functools
import inspect
import logging
import sys
import threading
import traceback

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Optional, Union
except ImportError:
    pass

# Pelix constants
import pelix.constants

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Using Python 3
PYTHON_3 = sys.version_info[0] == 3

# ------------------------------------------------------------------------------


@contextlib.contextmanager
def use_service(bundle_context, svc_reference):
    """
    Utility context to safely use a service in a "with" block.
    It looks after the the given service and releases its reference when
    exiting the context.

    :param bundle_context: The calling bundle context
    :param svc_reference: The reference of the service to use
    :return: The requested service
    :raise BundleException: Service not found
    :raise TypeError: Invalid service reference
    """
    if svc_reference is None:
        raise TypeError("Invalid ServiceReference")

    try:
        # Give the service
        yield bundle_context.get_service(svc_reference)
    finally:
        try:
            # Release it
            bundle_context.unget_service(svc_reference)
        except pelix.constants.BundleException:
            # Service might have already been unregistered
            pass


# ------------------------------------------------------------------------------


# Mimic ArgSpec from getargspec()
ArgSpec = collections.namedtuple("ArgSpec", "args varargs keywords defaults")

if hasattr(inspect, "signature"):
    # Python 3.3+
    def get_method_arguments(method):
        """
        inspect.signature()-based way to get the arguments of a method.

        :param method: The method to extract the signature from
        :return: The arguments specification, without self
        """
        # Use the recommended signature method
        signature = inspect.signature(method)

        args = []
        varargs = None
        keywords = None
        defaults = []

        for param in signature.parameters.values():
            kind = param.kind
            if kind == inspect.Parameter.VAR_POSITIONAL:
                varargs = param.name
            elif kind == inspect.Parameter.VAR_KEYWORD:
                keywords = param.name
            else:
                args.append(param.name)

            if param.default is not param.empty:
                defaults.append(param.default)

        return ArgSpec(args, varargs, keywords, defaults or None)


else:
    import types

    def get_method_arguments(method):
        """
        inspect.getargspec()-based way to get the position of arguments.

        The self argument is removed from the result.

        :param method: The method to extract the signature from
        :return: The arguments specification, without self
        """
        # pylint: disable=W1505
        arg_spec = inspect.getargspec(method)

        if not isinstance(method, types.FunctionType):
            # Filter out the "self" argument
            args = arg_spec.args[1:]
        else:
            args = arg_spec.args

        return ArgSpec(
            args, arg_spec.varargs, arg_spec.keywords, arg_spec.defaults
        )


# ------------------------------------------------------------------------------


class Deprecated(object):
    """
    Prints a warning when using the decorated method
    """

    def __init__(self, message=None, logger=None):
        """
        Sets the deprecation message, e.g. to indicate which method to call
        instead.
        If a logger is given, its 'warning' method will be called to print the
        message; else the standard 'print' method will be used.

        :param message: Message to be printed
        :param logger: The name of the logger to use, or None.
        """
        self.__message = message or "Deprecated method"
        self.__logger = logger or None
        self.__already_logged = False

    def __log(self, method_name):
        """
        Logs the deprecation message on first call, does nothing after

        :param method_name: Name of the deprecated method
        """
        if not self.__already_logged:
            # Print only if not already done
            stack = "\n\t".join(traceback.format_stack())

            logging.getLogger(self.__logger).warning(
                "%s: %s\n%s", method_name, self.__message, stack
            )
            self.__already_logged = True

    def __call__(self, method):
        """
        Applies the modifications

        :param method: The decorated method
        :return: The wrapped method
        """
        # Prepare the wrapped call
        @functools.wraps(method)
        def wrapped(*args, **kwargs):
            """
            Wrapped deprecated method
            """
            self.__log(method.__name__)
            return method(*args, **kwargs)

        return wrapped


# ------------------------------------------------------------------------------


class Synchronized(object):
    """
    A synchronizer for global methods
    """

    def __init__(self, lock=None):
        """
        Sets up the decorator. If 'lock' is None, an RLock() is created for
        this decorator.

        :param lock: The lock to be used for synchronization (can be None)
        """
        if not is_lock(lock):
            self.__lock = threading.RLock()
        else:
            self.__lock = lock

    def __call__(self, method):
        """
        Sets up the decorated method

        :param method: The decorated method
        :return: The wrapped method
        """

        @functools.wraps(method)
        def wrapped(*args, **kwargs):
            """
            The wrapping method
            """
            with self.__lock:
                return method(*args, **kwargs)

        return wrapped


def SynchronizedClassMethod(*locks_attr_names, **kwargs):
    # pylint: disable=C1801
    """
    A synchronizer decorator for class methods. An AttributeError can be raised
    at runtime if the given lock attribute doesn't exist or if it is None.

    If a parameter ``sorted`` is found in ``kwargs`` and its value is True,
    then the list of locks names will be sorted before locking.

    :param locks_attr_names: A list of the lock(s) attribute(s) name(s) to be
                             used for synchronization
    :return: The decorator method, surrounded with the lock
    """
    # Filter the names (remove empty ones)
    locks_attr_names = [
        lock_name for lock_name in locks_attr_names if lock_name
    ]

    if not locks_attr_names:
        raise ValueError("The lock names list can't be empty")

    if "sorted" not in kwargs or kwargs["sorted"]:
        # Sort the lock names if requested
        # (locking always in the same order reduces the risk of dead lock)
        locks_attr_names = list(locks_attr_names)
        locks_attr_names.sort()

    def wrapped(method):
        """
        The wrapping method

        :param method: The wrapped method
        :return: The wrapped method
        :raise AttributeError: The given attribute name doesn't exist
        """

        @functools.wraps(method)
        def synchronized(self, *args, **kwargs):
            """
            Calls the wrapped method with a lock
            """
            # Raises an AttributeError if needed
            locks = [getattr(self, attr_name) for attr_name in locks_attr_names]
            locked = collections.deque()
            i = 0

            try:
                # Lock
                for lock in locks:
                    if lock is None:
                        # No lock...
                        raise AttributeError(
                            "Lock '{0}' can't be None in class {1}".format(
                                locks_attr_names[i], type(self).__name__
                            )
                        )

                    # Get the lock
                    i += 1
                    lock.acquire()
                    locked.appendleft(lock)

                # Use the method
                return method(self, *args, **kwargs)

            finally:
                # Unlock what has been locked in all cases
                for lock in locked:
                    lock.release()

                locked.clear()
                del locks[:]

        return synchronized

    # Return the wrapped method
    return wrapped


def is_lock(lock):
    """
    Tests if the given lock is an instance of a lock class
    """
    if lock is None:
        # Don't do useless tests
        return False

    for attr in "acquire", "release", "__enter__", "__exit__":
        if not hasattr(lock, attr):
            # Missing something
            return False

    # Same API as a lock
    return True


# ------------------------------------------------------------------------------


def read_only_property(value):
    """
    Makes a read-only property that always returns the given value
    """
    return property(lambda cls: value)


# ------------------------------------------------------------------------------


def remove_all_occurrences(sequence, item):
    """
    Removes all occurrences of item in the given sequence

    :param sequence: The items list
    :param item: The item to be removed
    """
    if sequence is None:
        return

    while item in sequence:
        sequence.remove(item)


def remove_duplicates(items):
    """
    Returns a list without duplicates, keeping elements order

    :param items: A list of items
    :return: The list without duplicates, in the same order
    """
    if items is None:
        return items

    new_list = []
    for item in items:
        if item not in new_list:
            new_list.append(item)
    return new_list


# ------------------------------------------------------------------------------


def add_listener(registry, listener):
    """
    Adds a listener in the registry, if it is not yet in

    :param registry: A registry (a list)
    :param listener: The listener to register
    :return: True if the listener has been added
    """
    if listener is None or listener in registry:
        return False

    registry.append(listener)
    return True


def remove_listener(registry, listener):
    """
    Removes a listener from the registry

    :param registry: A registry (a list)
    :param listener: The listener to remove
    :return: True if the listener was in the list
    """
    if listener is not None and listener in registry:
        registry.remove(listener)
        return True

    return False


# ------------------------------------------------------------------------------


if PYTHON_3:
    # Python 3 interpreter : bytes & str
    def is_bytes(string):
        """
        Utility method to test if the given parameter is a string
        (Python 2.x) or a bytes (Python 3.x) object

        :param string: A potential string object
        :return: True if the given object is a bytes string
        """
        # str in Python 2 is bytes in Python 3
        return isinstance(string, bytes)

    def is_string(string):
        """
        Utility method to test if the given parameter is a string
        (Python 2.x, 3.x) or a unicode (Python 2.x) object

        :param string: A potential string object
        :return: True if the given object is a string object or a Python 2.x
                 unicode object
        """
        # Python 3 only have the str string type
        return isinstance(string, str)

    def to_bytes(data, encoding="UTF-8"):
        """
        Converts the given string to an array of bytes.
        Returns the first parameter if it is already an array of bytes.

        :param data: A unicode string
        :param encoding: The encoding of data
        :return: The corresponding array of bytes
        """
        if isinstance(data, bytes):
            # Nothing to do
            return data

        return data.encode(encoding)

    def to_str(data, encoding="UTF-8"):
        """
        Converts the given parameter to a string.
        Returns the first parameter if it is already an instance of ``str``.

        :param data: A string
        :param encoding: The encoding of data
        :return: The corresponding string
        """
        if isinstance(data, str):
            # Nothing to do
            return data

        return str(data, encoding)

    # Same operation
    # pylint: disable=C0103
    to_unicode = to_str

else:
    # Python 2 interpreter : str & unicode
    def is_bytes(string):
        """
        Utility method to test if the given parameter is a string
        (Python 2.x) or a bytes (Python 3.x) object

        :param string: A potential string object
        :return: True if the given object is a bytes string
        """
        # str in Python 2 is bytes in Python 3
        return isinstance(string, str)

    def is_string(string):
        """
        Utility method to test if the given parameter is a string
        (Python 2.x, 3.x) or a unicode (Python 2.x) object

        :param string: A potential string object
        :return: True if the given object is a string object or a Python 2.x
                 unicode object
        """
        # Python 2 also have unicode
        # pylint: disable=E0602
        return isinstance(string, (str, unicode))

    def to_str(data, encoding="UTF-8"):
        """
        Converts the given parameter to a string.
        Returns the first parameter if it is already an instance of ``str``.

        :param data: A string
        :param encoding: The encoding of data
        :return: The corresponding string
        """
        if type(data) is str:
            # Nothing to do
            return data

        return data.encode(encoding)

    # Same operation
    # pylint: disable=C0103
    to_bytes = to_str

    def to_unicode(data, encoding="UTF-8"):
        """
        Converts the given string to an unicode string using ``str.decode()``.
        Returns the first parameter if it is already an instance of
        ``unicode``.

        :param data: A string
        :param encoding: The encoding of data
        :return: The corresponding ``unicode`` string
        """
        # pylint: disable=E0602
        if type(data) is unicode:
            # Nothing to do
            return data

        return data.decode(encoding)


# ------------------------------------------------------------------------------


def to_iterable(value, allow_none=True):
    """
    Tries to convert the given value to an iterable, if necessary.
    If the given value is a list, a list is returned; if it is a string, a list
    containing one string is returned, ...

    :param value: Any object
    :param allow_none: If True, the method returns None if value is None, else
                       it returns an empty list
    :return: A list containing the given string, or the given value
    """
    if value is None:
        # None given
        if allow_none:
            return None

        return []

    elif isinstance(value, (list, tuple, set, frozenset)):
        # Iterable given, return it as-is
        return value

    # Return a one-value list
    return [value]


# ------------------------------------------------------------------------------


class EventData(object):
    """
    A threading event with some associated data
    """

    __slots__ = ("__event", "__data", "__exception")

    def __init__(self):
        """
        Sets up the event
        """
        self.__event = threading.Event()
        self.__data = None
        self.__exception = None

    @property
    def data(self):
        # type: () -> Any
        """
        Returns the associated value
        """
        return self.__data

    @property
    def exception(self):
        # type: () -> BaseException
        """
        Returns the exception used to stop the wait() method
        """
        return self.__exception

    def clear(self):
        """
        Clears the event
        """
        self.__event.clear()
        self.__data = None
        self.__exception = None

    def is_set(self):
        # type: () -> bool
        """
        Checks if the event is set
        """
        return self.__event.is_set()

    def set(self, data=None):
        # type: (Any) -> None
        """
        Sets the event
        """
        self.__data = data
        self.__exception = None
        self.__event.set()

    def raise_exception(self, exception):
        # type: (BaseException) -> None
        """
        Raises an exception in wait()

        :param exception: An Exception object
        """
        self.__data = None
        self.__exception = exception
        self.__event.set()

    def wait(self, timeout=None):
        # type: (Optional[int]) -> object
        """
        Waits for the event or for the timeout

        :param timeout: Wait timeout (in seconds)
        :return: True if the event as been set, else False
        """
        result = self.__event.wait(timeout)
        # pylint: disable=E0702
        # Pylint seems to miss the "is None" check below
        if self.__exception is None:
            return result
        else:
            raise self.__exception


class CountdownEvent(object):
    """
    Sets up an Event once the internal integer reaches 0
    (kind of the opposite of a semaphore)
    """

    def __init__(self, value):
        # type: (int) -> None
        """
        Sets up the counter

        :param value: The initial value of the counter, which must be greater
                      than 0.
        :raise ValueError: The value is not greater than 0
        """
        if value <= 0:
            raise ValueError("Initial value is not greater than 0")

        self.__lock = threading.Lock()
        self.__value = value
        self.__event = threading.Event()

    def is_set(self):
        # type: () -> bool
        """
        Checks if the event is set
        """
        return self.__event.is_set()

    def step(self):
        # type: () -> bool
        """
        Decreases the internal counter. Raises an error if the counter goes
        below 0

        :return: True if this step was the final one, else False
        :raise ValueError: The counter has gone below 0
        """
        with self.__lock:
            self.__value -= 1
            if self.__value == 0:
                # All done
                self.__event.set()
                return True
            elif self.__value < 0:
                # Gone too far
                raise ValueError("The counter has gone below 0")

        return False

    def wait(self, timeout=None):
        # type: (Optional[int]) -> bool
        """
        Waits for the event or for the timeout

        :param timeout: Wait timeout (in seconds)
        :return: True if the event as been set, else False
        """
        return self.__event.wait(timeout)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
LDAP filter parser tests

:author: Thomas Calmant
"""

import inspect

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from pelix.ldapfilter import get_ldap_filter
import pelix.ldapfilter

# ------------------------------------------------------------------------------

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


def applyTest(self, filters, key):
    """
    Applies a list of tests according to the given dictionary

    Dictionary format: filter -> ([True, results], [False, results])

    @param filters: A filters test dictionary
    @param key: The key to use in the property dictionary
    """
    props = {}

    for filter_str, tests in filters.items():
        ldap_filter = get_ldap_filter(filter_str)
        self.assertIsNotNone(ldap_filter, "{0} is a valid filter"
                             .format(filter_str))

        for good in tests[0]:
            props[key] = good
            self.assertTrue(ldap_filter.matches(props),
                            "Filter '{0}' should match {1}"
                            .format(ldap_filter, props))

        for bad in tests[1]:
            props[key] = bad
            self.assertFalse(ldap_filter.matches(props),
                             "Filter '{0}' should not match {1}"
                             .format(ldap_filter, props))

# ------------------------------------------------------------------------------


class LDAPUtilitiesTest(unittest.TestCase):
    """
    Tests for LDAP utility methods
    """
    def testComparator2str(self):
        """
        Tests comparator2str()
        """
        for comparator in ('=', '<', '<=', '>', '>=', '=', '~='):
            # Parse a criteria with that comparator
            ldap_filter = get_ldap_filter("(a{0}1)".format(comparator))

            # Get the string version of the parsed comparator
            str_comparator = pelix.ldapfilter.comparator2str(
                ldap_filter.comparator)

            self.assertEqual(str_comparator, comparator,
                             "Bad string for comparator '{0}': '{1}'"
                             .format(comparator, str_comparator))

        # Invalid comparators
        for comparator in (None, str, str(), int()):
            str_comparator = pelix.ldapfilter.comparator2str(comparator)
            self.assertEqual(str_comparator, "??",
                             "Bad string for comparator '{0}': '{1}'"
                             .format(comparator, str_comparator))

    def testOperator2str(self):
        """
        Tests operator2str()
        """
        operators = {pelix.ldapfilter.AND: "&",
                     pelix.ldapfilter.OR: "|",
                     pelix.ldapfilter.NOT: "!"}

        for operator, str_operator in operators.items():
            conv_operator = pelix.ldapfilter.operator2str(operator)
            self.assertEqual(str_operator, conv_operator,
                             "Invalid operator conversion '{0}': '{1}'"
                             .format(str_operator, conv_operator))

        for operator in (None, str, int, str(), "AND", "OR", "NOT", 42):
            conv_operator = pelix.ldapfilter.operator2str(operator)
            self.assertEqual("<unknown>", conv_operator,
                             "Invalid operator conversion '{0}': '{1}'"
                             .format(operator, conv_operator))

    def testEscapeLDAP(self):
        """
        Tests escape_LDAP() and unescape_LDAP()

        Tested values from :
        https://www.owasp.org/index.php/Preventing_LDAP_Injection_in_Java
        """
        # Tested values: normal -> escaped
        tested_values = {None: None,
                         # Empty string -> Empty string
                         "": "",
                         # No escape needed
                         "Helloé": "Helloé",
                         # Sharp escape
                         "# Helloé": "\\# Helloé",
                         # Space escapes
                         " Helloé": "\\ Helloé",
                         "Helloé ": "Helloé\\ ",
                         "Hello é": "Hello é",
                         # Only spaces
                         "   ": "\\  \\ ",
                         # Complex
                         ' Hello\\ + , "World" ; ':
                             '\\ Hello\\\\ \\+ \\, \\"World\\" \\;\\ '}

        for normal, escaped in tested_values.items():
            # Escape
            ldap_escape = pelix.ldapfilter.escape_LDAP(normal)
            self.assertEqual(escaped, ldap_escape,
                             "Invalid escape '{0}' should be '{1}'"
                             .format(ldap_escape, escaped))

            # Un-escape
            ldap_unescape = pelix.ldapfilter.unescape_LDAP(ldap_escape)
            self.assertEqual(escaped, ldap_escape,
                             "Invalid unescape '{0}' should be '{1}'"
                             .format(ldap_unescape, normal))

    def testParseCriteria(self):
        """
        Tests the _parse_ldap_criteria() method
        """
        # Invalid criteria / incomplete operators
        for invalid in ("test 12", "=12", "test=", "test~12", "test~"):
            self.assertRaises(ValueError,
                              pelix.ldapfilter._parse_ldap_criteria, invalid)

        # Escape test (avoid the first test)
        value = "a=2 tes\\ t\\==1\\ 2\\~"
        criteria = pelix.ldapfilter._parse_ldap_criteria(value, 4, len(value))
        self.assertEqual(criteria.name, "tes t=",
                         "Escaped name not correctly parsed")
        self.assertEqual(criteria.value, "1 2~",
                         "Escaped value not correctly parsed")

        # Invalid test range
        test_str = "test=1"
        self.assertRaises(ValueError,
                          pelix.ldapfilter._parse_ldap_criteria,
                          test_str, -1, len(test_str))

        self.assertRaises(ValueError,
                          pelix.ldapfilter._parse_ldap_criteria,
                          test_str, len(test_str) - 1, len(test_str) - 2)

    def testParseLDAP(self):
        """
        Tests the _parse_ldap method
        """
        # Empty filters
        for empty in (None, "", "   "):
            self.assertEqual(pelix.ldapfilter._parse_ldap(empty), None,
                             "Empty filter: must return None")

        # Invalid filters
        for invalid in ("(", "(test", "(|(test=True)(test=False))))",
                        "(test)", "((test=1)(test2=1))"):
            self.assertRaises(ValueError, pelix.ldapfilter._parse_ldap,
                              invalid)

        # Criteria parsing
        criteria = pelix.ldapfilter.LDAPCriteria(
            "test", True, pelix.ldapfilter._comparator_eq)
        self.assertEqual(pelix.ldapfilter._parse_ldap(str(criteria)),
                         criteria, "Incorrect result: {0}".format(criteria))

        # Filter parsing
        ldap_filter = pelix.ldapfilter.LDAPFilter(pelix.ldapfilter.AND)
        sub_filter = pelix.ldapfilter.LDAPFilter(pelix.ldapfilter.OR)
        sub_filter_2 = pelix.ldapfilter.LDAPFilter(pelix.ldapfilter.NOT)
        sub_filter_2.subfilters.append(criteria)
        sub_filter.subfilters.append(sub_filter_2)
        sub_filter.subfilters.append(criteria)
        ldap_filter.subfilters.append(sub_filter)
        ldap_filter.subfilters.append(criteria)

        criteria_2 = pelix.ldapfilter.LDAPCriteria(
            "te st=", True, pelix.ldapfilter._comparator_eq)
        ldap_filter.subfilters.append(criteria_2)

        self.assertEqual(pelix.ldapfilter._parse_ldap(str(ldap_filter)),
                         ldap_filter.normalize(),
                         "Incorrect result: {0}".format(ldap_filter))

    def testCombine(self):
        """
        Tests the combine_filters() method
        """
        # Standard case
        criterias = [get_ldap_filter("(test=True)"),
                     get_ldap_filter("(test2=False)")]

        for operator in (pelix.ldapfilter.AND, pelix.ldapfilter.OR):
            ldap_filter = pelix.ldapfilter.combine_filters(criterias, operator)

            assert isinstance(ldap_filter, pelix.ldapfilter.LDAPFilter)
            self.assertEqual(ldap_filter.operator, operator,
                             "Invalid operator")
            self.assertEqual(len(ldap_filter.subfilters), 2,
                             "Invalid count of sub filters")

            for criteria in criterias:
                self.assertIn(criteria, ldap_filter.subfilters,
                              "A criteria is missing in the result")

        # No filter given
        for empty in (None, [], tuple(), (None, None, None)):
            self.assertIsNone(pelix.ldapfilter.combine_filters(empty),
                              "Can't combine an empty list of filters")

        # Invalid types
        for invalid in ("Filters", get_ldap_filter("(!(test=True))")):
            self.assertRaises(TypeError, pelix.ldapfilter.combine_filters,
                              invalid)

        ldap_filter_1 = get_ldap_filter("(test=True)")
        ldap_filter_2 = pelix.ldapfilter.LDAPFilter(pelix.ldapfilter.AND)

        # Unique filter in result
        self.assertIs(pelix.ldapfilter.combine_filters((None, ldap_filter_1)),
                      ldap_filter_1, "The result of combine must be minimal")

        self.assertIs(pelix.ldapfilter.combine_filters((ldap_filter_1,
                                                        ldap_filter_2)),
                      ldap_filter_1, "The result of combine must be minimal")

    def testGetLdapFilter(self):
        """
        Tests the get_ldap_filter() method
        """
        # Simple parsing / re-parsing test
        for str_filter, filter_type in \
                (("(|(test=True)(test2=False))", pelix.ldapfilter.LDAPFilter),
                 ("(test=True)", pelix.ldapfilter.LDAPCriteria)):
            ldap_filter = get_ldap_filter(str_filter)
            assert isinstance(ldap_filter, filter_type)

            self.assertEqual(str(ldap_filter), str_filter, "Invalid parsing")
            self.assertIs(get_ldap_filter(ldap_filter), ldap_filter,
                          "get_ldap_filter should return the given object.")

        # Empty filters
        for empty in (None, "", "   "):
            self.assertIsNone(get_ldap_filter(empty),
                              "Empty filter should return None")

        # Invalid types
        for invalid in (1, [], {}):
            self.assertRaises(TypeError, get_ldap_filter, invalid)

# ------------------------------------------------------------------------------


class LDAPCriteriaTest(unittest.TestCase):
    """
    Tests for the LDAP criteria behavior
    """
    def testInit(self):
        """
        Tests __init__() behavior on invalid values
        """
        for name in (None, "", "name"):
            for value in (None, "", "value"):
                for comparator in (None, True, lambda x: True):
                    if not all((name, value, comparator)) \
                            or (not inspect.isfunction(comparator)
                                and not inspect.ismethod(comparator)):
                        # One value is None
                        self.assertRaises(ValueError,
                                          pelix.ldapfilter.LDAPCriteria, name,
                                          value, comparator)

                    else:
                        # All values are OK
                        criteria = pelix.ldapfilter.LDAPCriteria(name, value,
                                                                 comparator)
                        self.assertEqual(name, criteria.name,
                                         "Name modified")
                        self.assertEqual(value, criteria.value,
                                         "Value modified")
                        self.assertTrue(
                            inspect.ismethod(criteria.comparator)
                            or inspect.isfunction(criteria.comparator),
                            "Invalid comparator accepted")

    def testRepr(self):
        """
        Tests repr() -> eval() transformation
        """
        # String filter: no spaces between operators nor operands
        # => allows direct str() results tests
        str_criteria = "(test=False)"

        # Make the filter
        criteria = get_ldap_filter(str_criteria)
        assert isinstance(criteria, pelix.ldapfilter.LDAPCriteria)

        # Assert strings representations are equals
        self.assertEqual(str_criteria, str(criteria))

        # Conversion
        repr_criteria = repr(criteria)
        self.assertIn(str_criteria, repr_criteria,
                      "The representation must contain the criteria string")

        # Evaluation
        eval_filter = eval(repr_criteria)

        # Equality based on the string form
        self.assertEqual(str_criteria, str(eval_filter), "Invalid evaluation")

    def testEq(self):
        """
        Tests the LDAPFilter objects equality
        """
        # Some filter
        str_filter = "(&(test=False)(test2=True))"
        ldap_filter = get_ldap_filter(str_filter)

        # Test with other values
        self.assertNotEqual(ldap_filter, None, "Filter is not equal to None")
        self.assertEqual(ldap_filter, ldap_filter, "Filter is not self-equal")
        self.assertNotEqual(pelix.ldapfilter.LDAPFilter(pelix.ldapfilter.NOT),
                            pelix.ldapfilter.LDAPCriteria(
                                'test', 123,
                                pelix.ldapfilter._comparator_approximate),
                            "Invalid equality (type)")

        # Tests order must not provide a different filter
        str_filter_2 = "(&(test2 = True)(test = False))"
        ldap_filter_2 = get_ldap_filter(str_filter_2)
        self.assertEqual(ldap_filter, ldap_filter_2, "Filters are not equal")

        # Inequality check
        self.assertNotEqual(ldap_filter, get_ldap_filter("(test2=true)"),
                            "Invalid equality (type)")
        self.assertNotEqual(get_ldap_filter("(test2=true)"), ldap_filter,
                            "Invalid equality (type, reverse)")

        self.assertNotEqual(
            ldap_filter,
            get_ldap_filter("(&(test=False)(test2=True)(test3=1))"),
            "Invalid equality (size)")
        self.assertNotEqual(
            get_ldap_filter("(&(test=False)(test2=True)(test3=1))"),
            ldap_filter, "Invalid equality (size, reverse)")

        self.assertNotEqual(ldap_filter,
                            get_ldap_filter("(&(test1=False)(test2=True))"),
                            "Invalid equality (sub-filter)")
        self.assertNotEqual(get_ldap_filter("(&(test1=False)(test2=True))"),
                            ldap_filter,
                            "Invalid equality (sub-filter, reverse)")

        self.assertNotEqual(ldap_filter,
                            get_ldap_filter("(|(test=False)(test2=True))"),
                            "Invalid equality (operator)")
        self.assertNotEqual(
            get_ldap_filter("(|(test=False)(test2=True))"),
            ldap_filter, "Invalid equality (operator, reverse)")

    def testNormalize(self):
        """
        Tests the normalize() method
        """
        criteria = get_ldap_filter("(test=True)")
        self.assertIs(criteria, criteria.normalize(),
                      "Criteria.normalize() must return itself")

    def testEmptyCriteria(self):
        """
        Empty filter test
        """
        self.assertIsNone(get_ldap_filter(None),
                          "None filter must return None")
        self.assertIsNone(get_ldap_filter(""),
                          "Empty filter must return None")
        self.assertIsNone(get_ldap_filter(" "),
                          "Trimmed filter must return None")

    def testSimpleCriteria(self):
        """
        Simple boolean filter test
        """
        props = {}

        ldap_filter = get_ldap_filter("(valid=True)")
        self.assertIsNotNone(ldap_filter, "Filter should not be None")

        # Test with a single property
        props["valid"] = True
        self.assertTrue(ldap_filter.matches(props),
                        "Filter '{0}' should match {1}"
                        .format(ldap_filter, props))

        props["valid"] = False
        self.assertFalse(ldap_filter.matches(props),
                         "Filter '{0}' should not match {1}"
                         .format(ldap_filter, props))

        # Test the ignorance of other properties
        props["valid2"] = True
        self.assertFalse(ldap_filter.matches(props),
                         "Filter '{0}' should not match {1}"
                         .format(ldap_filter, props))

        props["valid"] = "True"
        self.assertTrue(ldap_filter.matches(props),
                        "Filter '{0}' should match {1}"
                        .format(ldap_filter, props))

    def testPresenceCriteria(self):
        """
        Test the presence filter
        """
        props = {}

        ldap_filter = get_ldap_filter("(valid=*)")
        self.assertIsNotNone(ldap_filter, "Filter should not be None")

        # Missing value
        self.assertFalse(ldap_filter.matches(props),
                         "Filter '{0}' should not match {1}"
                         .format(ldap_filter, props))

        # Still missing
        props["valid2"] = True
        self.assertFalse(ldap_filter.matches(props),
                         "Filter '{0}' should not match {1}"
                         .format(ldap_filter, props))

        # Value present
        props["valid"] = True
        self.assertTrue(ldap_filter.matches(props),
                        "Filter '{0}' should match {1}"
                        .format(ldap_filter, props))

        props["valid"] = False
        self.assertTrue(ldap_filter.matches(props),
                        "Filter '{0}' should match {1}"
                        .format(ldap_filter, props))

        # Some other type
        props["valid"] = "1234"
        self.assertTrue(ldap_filter.matches(props),
                        "Filter '{0}' should match {1}"
                        .format(ldap_filter, props))

        # Empty values
        for empty in ('', [], tuple()):
            props["valid"] = empty
            self.assertFalse(ldap_filter.matches(props),
                             "Filter '{0}' should not match {1}"
                             .format(ldap_filter, props))

    def testStarCriteria(self):
        """
        Tests the star/joker filter on strings
        """
        filters = {}
        # Simple string test
        filters["(string=after*)"] = (("after", "after1234"),
                                      ("1324after", "before", "After"))

        filters["(string=*before)"] = (("before", "1234before"),
                                       ("after", "before1234"), "Before")

        filters["(string=*middle*)"] = (("middle", "aaamiddle1234",
                                         "middle456", "798middle"),
                                        ("miDDle",))

        filters["(string=*mi*ed*)"] = (("mixed", "mixed1234", "798mixed",
                                        "mi O_O ed"), ("Mixed",))

        # List test
        filters["(string=*li*ed*)"] = ((["listed"], ["toto", "aaaliXed123"]),
                                       ([], ["LixeD"], ["toto"]))

        applyTest(self, filters, "string")

        # Direct call test
        self.assertFalse(pelix.ldapfilter._comparator_star('T*ue', True),
                         "String star can't be compared to other values")

        self.assertTrue(pelix.ldapfilter._comparator_star('T*ue', 'True'),
                        "String star test failure")

    def testListCriteria(self):
        """
        Test the presence filter on lists
        """
        filters = {}
        filters["(list=toto)"] = ((["toto"], ["titi", "toto"],
                                   ["toto", "titi"],
                                   ["titi", "toto", "tutu"]),
                                  ([], ["titi"], ["*toto*"]))

        applyTest(self, filters, "list")

    def testInequalityCriteria(self):
        """
        Test the inequality operators
        """
        filters = {}
        filters["(id<10)"] = (("-10", -10, 0, 9), (10, 11, "12"))
        filters["(id<=10)"] = (("-10", -10, 0, 9, 10), (11, "12"))
        filters["(id>=10)"] = ((10, 11, "12"), ("-10", -10, 0, 9))
        filters["(id>10)"] = ((11, "12"), ("-10", -10, 0, 9, 10))

        applyTest(self, filters, "id")

        # Direct call test (for other cases)
        self.assertTrue(pelix.ldapfilter._comparator_gt('13', 14.0),
                        "Integer/Float comparison error")
        self.assertTrue(pelix.ldapfilter._comparator_gt('13.0', 14.0),
                        "Float/Float comparison error")
        self.assertTrue(pelix.ldapfilter._comparator_gt('13.0', 14),
                        "Float/Integer comparison error")

        self.assertTrue(pelix.ldapfilter._comparator_lt('13', 12.0),
                        "Integer/Float comparison error")
        self.assertTrue(pelix.ldapfilter._comparator_lt('13.0', 12.0),
                        "Float/Float comparison error")
        self.assertTrue(pelix.ldapfilter._comparator_lt('13.0', 12),
                        "Float/Integer comparison error")

        self.assertFalse(pelix.ldapfilter._comparator_gt('13.0', 14 + 1j),
                         "Float/Complex comparison error")
        self.assertFalse(pelix.ldapfilter._comparator_gt('13', 14 + 1j),
                         "Integer/Complex comparison error")
        self.assertFalse(pelix.ldapfilter._comparator_gt('13.0 + 1j', 14),
                         "Complex/Integer comparison error")
        self.assertFalse(pelix.ldapfilter._comparator_gt('13.0 + 1j', 14.0),
                         "Complex/Float comparison error")

        self.assertFalse(pelix.ldapfilter._comparator_lt('13.0', 12 + 1j),
                         "Float/Complex comparison error")
        self.assertFalse(pelix.ldapfilter._comparator_lt('13', 12 + 1j),
                         "Integer/Complex comparison error")
        self.assertFalse(pelix.ldapfilter._comparator_lt('13.0 + 1j', 12),
                         "Complex/Integer comparison error")
        self.assertFalse(pelix.ldapfilter._comparator_lt('13.0 + 1j', 12.0),
                         "Complex/Float comparison error")

    def testApproximateCriteria(self):
        """
        Tests the approximate criteria
        """
        filters = {}

        # Simple string test
        filters["(string~=aBc)"] = (("abc", "ABC", "aBc", "Abc"),
                                    ("bac", "aDc"))

        # Simple list test
        filters["(string~=dEf)"] = ((["abc", "def"], ["DEF"]),
                                    ([], ["aBc"]))

        # Star test
        filters["(string~=*test*)"] = ((["bigTest", "def"], "test", "TEST42"),
                                       ([], ["aBc"], "T3st"))

        applyTest(self, filters, "string")

        # Direct call test (for other cases)
        self.assertFalse(
            pelix.ldapfilter._comparator_approximate('test', None),
            "Invalid None test result")

        self.assertTrue(
            pelix.ldapfilter._comparator_approximate('test', ['Test', 12]),
            "Invalid list test result")

# ------------------------------------------------------------------------------


class LDAPFilterTest(unittest.TestCase):
    """
    Tests for the LDAP filter behavior
    """
    def testInit(self):
        """
        Tests __init__() behavior on invalid values
        """
        # WARNING: don't test True, as True == 1 in Python
        for value in (None, '', -1, {}, lambda x: True):
            self.assertRaises(ValueError, pelix.ldapfilter.LDAPFilter, value)

        for operator in (pelix.ldapfilter.AND, pelix.ldapfilter.OR,
                         pelix.ldapfilter.NOT):
            ldap_filter = pelix.ldapfilter.LDAPFilter(operator)
            self.assertEqual(ldap_filter.operator, operator,
                             "Operator modified")
            self.assertEqual(len(ldap_filter.subfilters), 0,
                             "Filter not empty after init")

    def testRepr(self):
        """
        Test repr() -> eval() transformation
        """
        # String filter: no spaces between operators nor operands
        # => allows direct str() results tests
        str_filter = "(&(test=False)(test2=True))"

        # Make the filter
        ldap_filter = get_ldap_filter(str_filter)

        # Assert strings representations are equals
        self.assertEqual(str_filter, str(ldap_filter))

        # Conversion
        repr_filter = repr(ldap_filter)
        self.assertIn(str_filter, repr_filter,
                      "The representation must contain the filter string")

        # Evaluation
        eval_filter = eval(repr_filter)

        # Equality based on the string form
        self.assertEqual(str_filter, str(eval_filter), "Invalid evaluation")

        # Match test
        for test_value in (True, False):
            for test2_value in (True, False):
                for test3_value in (None, True, False, 42, "string"):
                    properties = {"test": test_value, "test2": test2_value,
                                  "test3": test3_value}

                    self.assertEqual(ldap_filter.matches(properties),
                                     eval_filter.matches(properties),
                                     "Different result found for {0}"
                                     .format(properties))

    def testEq(self):
        """
        Tests the LDAPFilter objects equality
        """
        # Some filter
        str_filter = "(&(test=False)(test2=True))"
        ldap_filter = get_ldap_filter(str_filter)

        # Test with other values
        self.assertNotEqual(ldap_filter, None, "Filter is not equal to None")
        self.assertEqual(ldap_filter, ldap_filter, "Filter is not self-equal")
        self.assertNotEqual(
            pelix.ldapfilter.LDAPFilter(pelix.ldapfilter.NOT),
            pelix.ldapfilter.LDAPCriteria(
                'test', 123, pelix.ldapfilter._comparator_approximate),
            "Invalid equality (type)")

        # Tests order must not provide a different filter
        str_filter_2 = "(&(test2=True)(test=False))"
        ldap_filter_2 = get_ldap_filter(str_filter_2)
        self.assertEqual(ldap_filter, ldap_filter_2, "Filters are not equal")

        # Inequality check
        self.assertNotEqual(ldap_filter, get_ldap_filter("(test2=true)"),
                            "Invalid equality (type)")

        self.assertNotEqual(
            ldap_filter,
            get_ldap_filter("(&(test=False)(test2=True)(test3=1))"),
            "Invalid equality (size)")

        self.assertNotEqual(ldap_filter,
                            get_ldap_filter("(&(test1=False)(test2=True))"),
                            "Invalid equality (sub-filter)")

        self.assertNotEqual(ldap_filter,
                            get_ldap_filter("(|(test=False)(test2=True))"),
                            "Invalid equality (operator)")

    def testAppend(self):
        """
        Tests the filter append() method
        """
        # "And" operator
        ldap_filter = pelix.ldapfilter.LDAPFilter(pelix.ldapfilter.AND)

        # Add a "normal" value
        criteria = get_ldap_filter("(test=True)")
        ldap_filter.append(criteria)
        self.assertEqual(len(ldap_filter.subfilters), 1, "Criteria not added")

        # Add invalid values
        for invalid_value in (None, "(test=False)",
                              "(|(test=True)(test2=False))"):
            self.assertRaises(TypeError, ldap_filter.append, invalid_value)

        # Special case: 'Not' operator
        ldap_filter = pelix.ldapfilter.LDAPFilter(pelix.ldapfilter.NOT)
        ldap_filter.append(criteria)

        # No more value must be accepted
        self.assertRaises(ValueError, ldap_filter.append, criteria)

    def testNormalize(self):
        """
        Tests the normalize() method
        """
        # Empty filter
        for operator in (pelix.ldapfilter.AND, pelix.ldapfilter.OR,
                         pelix.ldapfilter.NOT):
            self.assertEqual(pelix.ldapfilter.LDAPFilter(operator).normalize(),
                             None, "Empty filter normalized form must be None")

        # Standard filter
        ldap_filter = get_ldap_filter("(|(test=True)(test2=False))")
        self.assertIs(ldap_filter.normalize(), ldap_filter,
                      "Normalized filter must return itself")

        criteria = get_ldap_filter("(test=True)")

        # 'Not' Filter with 1 child
        ldap_filter = pelix.ldapfilter.LDAPFilter(pelix.ldapfilter.NOT)
        ldap_filter.append(criteria)
        self.assertIs(ldap_filter.normalize(), ldap_filter,
                      "'Not' filter with 1 child must return itself")

        # 'And', 'Or' filter
        for operator in (pelix.ldapfilter.AND, pelix.ldapfilter.OR):
            ldap_filter = pelix.ldapfilter.LDAPFilter(operator)
            ldap_filter.append(criteria)
            self.assertEqual(
                ldap_filter.normalize(), criteria,
                "'And' or 'Or' with 1 child must return the child")

    def testNot(self):
        """
        Tests the NOT operator
        """
        filters = {}

        filters["(test=False)"] = ((False, [False], [True, False]),
                                   (True, [True], "1123", 1, 0))

        filters["(!(test=False))"] = ((True, [True], "1123", 1, 0),
                                      (False, [False], [True, False]))

        # Simple cases
        applyTest(self, filters, "test")

        # NOT handles only one operand
        self.assertRaises(ValueError, get_ldap_filter,
                          "(!(test=True)(test2=False))")

    def testAnd(self):
        """
        Tests the AND operator
        """
        props = {}
        ldap_filter = get_ldap_filter("(&(test=True)(test2=False))")

        # Valid
        props["test"] = True
        props["test2"] = False
        self.assertTrue(ldap_filter.matches(props),
                        "Filter '{0}' should match {1}"
                        .format(ldap_filter, props))

        # Invalid...
        props["test"] = False
        props["test2"] = False
        self.assertFalse(ldap_filter.matches(props),
                         "Filter '{0}' should not match {1}"
                         .format(ldap_filter, props))

        props["test"] = False
        props["test2"] = True
        self.assertFalse(ldap_filter.matches(props),
                         "Filter '{0}' should not match {1}"
                         .format(ldap_filter, props))

        props["test"] = True
        props["test2"] = True
        self.assertFalse(ldap_filter.matches(props),
                         "Filter '{0}' should not match {1}"
                         .format(ldap_filter, props))

    def testOr(self):
        """
        Tests the OR operator
        """
        props = {}
        ldap_filter = get_ldap_filter("(|(test=True)(test2=False))")

        # Valid ...
        props["test"] = True
        props["test2"] = False
        self.assertTrue(ldap_filter.matches(props),
                        "Filter '{0}' should match {1}"
                        .format(ldap_filter, props))

        props["test"] = False
        props["test2"] = False
        self.assertTrue(ldap_filter.matches(props),
                        "Filter '{0}' should match {1}"
                        .format(ldap_filter, props))

        props["test"] = True
        props["test2"] = True
        self.assertTrue(ldap_filter.matches(props),
                        "Filter '{0}' should match {1}"
                        .format(ldap_filter, props))

        # Invalid...
        props["test"] = False
        props["test2"] = True
        self.assertFalse(ldap_filter.matches(props),
                         "Filter '{0}' should not match {1}"
                         .format(ldap_filter, props))

# ------------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix Remote Services: Java-compatible RPC, based on the Jabsorb library

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging

# JSON-RPC
from jsonrpclib.SimpleJSONRPCServer import (
    SimpleJSONRPCDispatcher,
    NoMulticallResult,
)
import jsonrpclib.jsonrpc as jsonrpclib

# Jabsorb conversion
import pelix.misc.jabsorb as jabsorb

# iPOPO Decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Validate,
    Invalidate,
    Requires,
    Property,
)

# Pelix constants
from pelix.utilities import to_str
import pelix.http
import pelix.remote
import pelix.remote.transport.commons as commons

# ------------------------------------------------------------------------------

# Documentation strings format
__docformat__ = "restructuredtext en"

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# ------------------------------------------------------------------------------

JABSORB_CONFIG = "ecf.jabsorb"
""" Remote Service configuration constant """

PROP_JABSORB_ENDPOINT_NAME = "{0}.name".format(JABSORB_CONFIG)
""" Name of the endpoint """

PROP_HTTP_ACCESSES = "{0}.accesses".format(JABSORB_CONFIG)
""" HTTP accesses (comma-separated String) """

HOST_SERVLET_PATH = "/JABSORB-RPC"
""" Default servlet path """

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


class _JabsorbRpcServlet(SimpleJSONRPCDispatcher):
    """
    A JSON-RPC servlet, replacing the SimpleJSONRPCDispatcher from jsonrpclib,
    converting data from and to Jabsorb format.
    """

    def __init__(self, dispatch_method, encoding=None):
        """
        Sets up the servlet
        """
        SimpleJSONRPCDispatcher.__init__(self, encoding)

        # Register the system.* functions
        self.register_introspection_functions()

        # Make a link to the dispatch method
        self._dispatch_method = dispatch_method

    def _simple_dispatch(self, name, params):
        """
        Dispatch method
        """
        try:
            # Internal method
            func = self.funcs[name]
        except KeyError:
            # Other method
            pass
        else:
            # Internal method found
            if isinstance(params, (list, tuple)):
                return func(*params)

            return func(**params)

        # Avoid calling this method in the "except" block, as it would be in
        # an exception state (logs will consider the KeyError as a failure)
        return self._dispatch_method(name, params)

    def do_POST(self, request, response):
        # pylint: disable=C0103
        """
        Handle a POST request

        :param request: The HTTP request bean
        :param response: The HTTP response handler
        """
        # Get the request JSON content
        data = jsonrpclib.loads(to_str(request.read_data()))

        # Convert from Jabsorb
        data = jabsorb.from_jabsorb(data)

        # Dispatch
        try:
            result = self._unmarshaled_dispatch(data, self._simple_dispatch)
        except NoMulticallResult:
            # No result (never happens, but who knows...)
            result = None

        if result is not None:
            # Convert result to Jabsorb
            if "result" in result:
                result["result"] = jabsorb.to_jabsorb(result["result"])

            # Store JSON
            result = jsonrpclib.jdumps(result)
        else:
            # It was a notification
            result = ""

        # Send the result
        response.send_content(200, result, "application/json-rpc")


# ------------------------------------------------------------------------------


@ComponentFactory(pelix.remote.FACTORY_TRANSPORT_JABSORBRPC_EXPORTER)
@Provides(pelix.remote.SERVICE_EXPORT_PROVIDER)
@Requires("_http", pelix.http.HTTP_SERVICE)
@Property("_path", pelix.http.HTTP_SERVLET_PATH, HOST_SERVLET_PATH)
@Property(
    "_kinds",
    pelix.remote.PROP_REMOTE_CONFIGS_SUPPORTED,
    (JABSORB_CONFIG, "jabsorb-rpc"),
)
class JabsorbRpcServiceExporter(commons.AbstractRpcServiceExporter):
    """
    JABSORB-RPC Remote Services exporter
    """

    def __init__(self):
        """
        Sets up the exporter
        """
        # Call parent
        super(JabsorbRpcServiceExporter, self).__init__()

        # Supported configurations
        self._kinds = None

        # HTTP Service
        self._http = None
        self._path = None

        # JSON-RPC servlet
        self._servlet = None

    def make_endpoint_properties(self, svc_ref, name, fw_uid):
        """
        Prepare properties for the ExportEndpoint to be created

        :param svc_ref: Service reference
        :param name: Endpoint name
        :param fw_uid: Framework UID
        :return: A dictionary of extra endpoint properties
        """
        # Get the Jabsorb-RPC endpoint name
        jabsorb_name = svc_ref.get_property(PROP_JABSORB_ENDPOINT_NAME)
        if jabsorb_name:
            # The end point name has been configured in the Jabsorb way
            name = jabsorb_name

        return {
            pelix.remote.PROP_ENDPOINT_NAME: name,
            # Jabsorb-RPC endpoint name
            PROP_JABSORB_ENDPOINT_NAME: name,
            # HTTP accesses, as a comma-separated string
            PROP_HTTP_ACCESSES: self.get_accesses(),
        }

    def get_accesses(self):
        """
        Retrieves the URLs to access this component as a comma-separated list.
        The first URL contains a '{server}' variable
        """
        # Get HTTP server access
        host, port = self._http.get_access()
        if ":" in host:
            # IPv6 address
            host = "[{0}]".format(host)

        # Return two accesses: with a {server} variable and with the
        # bound address
        model = "http{2}://{{server}}:{0}{1}".format(
            port, self._path, "s" if self._http.is_https() else ""
        )
        return ",".join((model, model.format(server=host)))

    @Validate
    def validate(self, context):
        """
        Component validated
        """
        # Call parent
        super(JabsorbRpcServiceExporter, self).validate(context)

        # Create/register the servlet
        self._servlet = _JabsorbRpcServlet(self.dispatch)
        self._http.register_servlet(self._path, self._servlet)

    @Invalidate
    def invalidate(self, context):
        """
        Component invalidated
        """
        # Unregister the servlet
        self._http.unregister(None, self._servlet)

        # Call parent
        super(JabsorbRpcServiceExporter, self).invalidate(context)

        # Clean up members
        self._servlet = None


# ------------------------------------------------------------------------------


class _ServiceCallProxy(object):
    # pylint: disable=R0903
    """
    Service call proxy
    """

    def __init__(self, name, url):
        """
        Sets up the call proxy

        :param name: End point name
        :param url: End point URL
        """
        self.__name = name
        self.__url = url

    def __getattr__(self, name):
        """
        Prefixes the requested attribute name by the endpoint name
        """
        # Make a proxy for this call
        # This is an ugly trick to handle multithreaded calls, as the
        # underlying proxy re-uses the same connection when possible: sometimes
        # it means sending a request before retrieving a result
        proxy = jsonrpclib.ServerProxy(self.__url)

        def wrapped_call(*args, **kwargs):
            """
            Wrapped call
            """
            # Get the method from the proxy
            method = getattr(proxy, "{0}.{1}".format(self.__name, name))

            # Convert arguments
            args = [jabsorb.to_jabsorb(arg) for arg in args]
            kwargs = {
                key: jabsorb.to_jabsorb(value) for key, value in kwargs.items()
            }

            result = method(*args, **kwargs)
            return jabsorb.from_jabsorb(result)

        return wrapped_call


# ------------------------------------------------------------------------------


@ComponentFactory(pelix.remote.FACTORY_TRANSPORT_JABSORBRPC_IMPORTER)
@Provides(pelix.remote.SERVICE_IMPORT_ENDPOINT_LISTENER)
@Property(
    "_kinds",
    pelix.remote.PROP_REMOTE_CONFIGS_SUPPORTED,
    (JABSORB_CONFIG, "jabsorb-rpc"),
)
class JabsorbRpcServiceImporter(commons.AbstractRpcServiceImporter):
    """
    JABSORB-RPC Remote Services importer
    """

    def __init__(self):
        """
        Sets up the exporter
        """
        # Call parent
        super(JabsorbRpcServiceImporter, self).__init__()

        # Component properties
        self._kinds = None

    def make_service_proxy(self, endpoint):
        """
        Creates the proxy for the given ImportEndpoint

        :param endpoint: An ImportEndpoint bean
        :return: A service proxy
        """
        # Get the access URL
        access_url = endpoint.properties.get(PROP_HTTP_ACCESSES)
        if not access_url:
            # No URL information
            _logger.warning("No access URL given: %s", endpoint)
            return None

        # Get the first URL in the list
        access_url = access_url.split(",")[0]

        if endpoint.server is not None:
            # Server information given
            access_url = access_url.format(server=endpoint.server)
        else:
            # Use the local IP as the source server, just in case
            local_server = "localhost"
            access_url = access_url.format(server=local_server)

        # Compute the name
        name = endpoint.properties.get(PROP_JABSORB_ENDPOINT_NAME)
        if not name:
            _logger.error("Remote endpoint has no name: %s", endpoint)
            return None

        # Prepare the proxy
        return _ServiceCallProxy(name, access_url)

    def clear_service_proxy(self, endpoint):
        """
        Destroys the proxy made for the given ImportEndpoint

        :param endpoint: An ImportEndpoint bean
        """
        # Nothing to do
        return



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Defines some iPOPO constants

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import contextlib

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

    # Pelix
    from pelix.framework import BundleContext, ServiceReference
except ImportError:
    pass

from pelix.framework import BundleException

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

SERVICE_IPOPO = "pelix.ipopo.core"
""" iPOPO service specification """

IPOPO_SERVICE_SPECIFICATION = SERVICE_IPOPO
""" Compatibility constant """

SERVICE_IPOPO_WAITING_LIST = "pelix.ipopo.waiting_list"
""" iPOPO waiting list service specification """

# ------------------------------------------------------------------------------

HANDLER_REQUIRES = "ipopo.requires"
""" The @Requires handler ID """

HANDLER_REQUIRES_BEST = "ipopo.requires.best"
""" The @RequiresBest handler ID """

HANDLER_REQUIRES_BRODCAST = "ipopo.requires.broadcast"
""" The @RequiresBroadcast handler ID """

HANDLER_REQUIRES_MAP = "ipopo.requires.map"
""" The @RequiresMap handler ID """

HANDLER_REQUIRES_VARIABLE_FILTER = "ipopo.requires.variable_filter"
""" The @RequiresVarFilter handler ID """

HANDLER_TEMPORAL = "ipopo.temporal"
""" The @Temporal handler ID """

HANDLER_PROVIDES = "ipopo.provides"
""" The @Provides handler ID """

HANDLER_PROPERTY = "ipopo.properties"
""" The @Property handler ID """

# ------------------------------------------------------------------------------

# Injected class fields
IPOPO_METHOD_CALLBACKS = "__ipopo_callbacks__"
""" Contains the list of callback types this method is decorated for """

IPOPO_METHOD_FIELD_CALLBACKS = "__ipopo_field_callbacks__"
""" Contains a list of tuples (field, callback type) """

IPOPO_FACTORY_CONTEXT = "__ipopo_factory_context__"
""" Storage of the FactoryContext object """

# Method called by the injected property (must be injected in the instance)
IPOPO_GETTER_SUFFIX = "_getter"
IPOPO_SETTER_SUFFIX = "_setter"
IPOPO_PROPERTY_PREFIX = "_ipopo_property"
IPOPO_HIDDEN_PROPERTY_PREFIX = "_ipopo_hidden_property"
IPOPO_CONTROLLER_PREFIX = "_ipopo_controller"

# Other injected information
IPOPO_VALIDATE_ARGS = "__ipopo_validate_args__"
""" Storage of the arguments for ``@ValidateComponent`` """

ARG_BUNDLE_CONTEXT = "bundle_context"
""" Represents the bundle context argument in ``@ValidateContext`` """

ARG_COMPONENT_CONTEXT = "component_context"
""" Represents the component context argument in ``@ValidateContext`` """

ARG_PROPERTIES = "properties"
""" Represents the component properties argument in ``@ValidateContext`` """

# ------------------------------------------------------------------------------

# Callbacks
IPOPO_CALLBACK_BIND = "BIND"
""" Bind: called when a dependency is injected """

IPOPO_CALLBACK_BIND_FIELD = "BIND_FIELD"
""" BindField: called when a dependency is injected in the given field """

IPOPO_CALLBACK_UPDATE = "UPDATE"
"""
Update: called when the properties of an injected dependency have been updated
"""

IPOPO_CALLBACK_UPDATE_FIELD = "UPDATE_FIELD"
"""
UpdateField: called when the properties of a dependency injected in the given
field have been updated
"""

IPOPO_CALLBACK_UNBIND = "UNBIND"
""" Unbind: called when a dependency is about to be removed """

IPOPO_CALLBACK_UNBIND_FIELD = "UNBIND_FIELD"
"""
UnbindField: called when a dependency is about to be removed from the given
field
"""

IPOPO_CALLBACK_VALIDATE = "VALIDATE"
"""
ValidateComponent: Called once all mandatory dependencies have been bound,
with component-specific parameters
"""

IPOPO_CALLBACK_INVALIDATE = "INVALIDATE"
"""
InvalidateComponent: Called when one the mandatory dependencies is unbound,
with component-specific parameters
"""

IPOPO_CALLBACK_POST_REGISTRATION = "POST_REGISTRATION"
"""
Post-Registration: called when a service of the component has been registered
"""

IPOPO_CALLBACK_POST_UNREGISTRATION = "POST_UNREGISTRATION"
"""
Post-Unregistration: called when a service of the component has been
unregistered.
"""

# Properties
IPOPO_INSTANCE_NAME = "instance.name"
""" Name of the component instance """

IPOPO_REQUIRES_FILTERS = "requires.filters"
""" Dictionary (field → filter) to override @Requires filters """

IPOPO_TEMPORAL_TIMEOUTS = "temporal.timeouts"
""" Dictionary (field → timeout) to override @Temporal timeouts """

IPOPO_AUTO_RESTART = "pelix.ipopo.auto_restart"
"""
If True, the component will be re-instantiated after its bundle has been
updated
"""

# ------------------------------------------------------------------------------


def get_ipopo_svc_ref(bundle_context):
    # type: (BundleContext) -> Optional[Tuple[ServiceReference, Any]]
    """
    Retrieves a tuple containing the service reference to iPOPO and the service
    itself

    :param bundle_context: The calling bundle context
    :return: The reference to the iPOPO service and the service itself,
             None if not available
    """
    # Look after the service
    ref = bundle_context.get_service_reference(SERVICE_IPOPO)
    if ref is None:
        return None

    try:
        # Get it
        svc = bundle_context.get_service(ref)
    except BundleException:
        # Service reference has been invalidated
        return None

    # Return both the reference (to call unget_service()) and the service
    return ref, svc


@contextlib.contextmanager
def use_ipopo(bundle_context):
    # type: (BundleContext) -> Any
    """
    Utility context to use the iPOPO service safely in a "with" block.
    It looks after the the iPOPO service and releases its reference when
    exiting the context.

    :param bundle_context: The calling bundle context
    :return: The iPOPO service
    :raise BundleException: Service not found
    """
    # Get the service and its reference
    ref_svc = get_ipopo_svc_ref(bundle_context)
    if ref_svc is None:
        raise BundleException("No iPOPO service available")

    try:
        # Give the service
        yield ref_svc[1]
    finally:
        try:
            # Release it
            bundle_context.unget_service(ref_svc[0])
        except BundleException:
            # Service might have already been unregistered
            pass


@contextlib.contextmanager
def use_waiting_list(bundle_context):
    # type: (BundleContext) -> Any
    """
    Utility context to use the iPOPO waiting list safely in a "with" block.
    It looks after the the iPOPO waiting list service and releases its
    reference when exiting the context.

    :param bundle_context: The calling bundle context
    :return: The iPOPO waiting list service
    :raise BundleException: Service not found
    """
    # Get the service and its reference
    ref = bundle_context.get_service_reference(SERVICE_IPOPO_WAITING_LIST)
    if ref is None:
        raise BundleException("No iPOPO waiting list service available")

    try:
        # Give the service
        yield bundle_context.get_service(ref)
    finally:
        try:
            # Release it
            bundle_context.unget_service(ref)
        except BundleException:
            # Service might have already been unregistered
            pass


# ------------------------------------------------------------------------------


class IPopoEvent(object):
    """
    An iPOPO event descriptor.
    """

    REGISTERED = 1
    """ A component factory has been registered """

    INSTANTIATED = 2
    """ A component has been instantiated, but not yet validated """

    VALIDATED = 3
    """ A component has been validated """

    INVALIDATED = 4
    """ A component has been invalidated """

    BOUND = 5
    """ A reference has been injected in the component """

    UNBOUND = 6
    """ A reference has been removed from the component """

    KILLED = 9
    """ A component has been killed (removed from the list of instances) """

    UNREGISTERED = 10
    """ A component factory has been unregistered """

    def __init__(self, kind, factory_name, component_name):
        # type: (int, str, Optional[str]) -> None
        """
        Sets up the iPOPO event

        :param kind: Kind of event
        :param factory_name: Name of the factory associated to the event
        :param component_name: Name of the component instance associated to the
                               event
        """
        self.__kind = kind
        self.__factory_name = factory_name
        self.__component_name = component_name

    def get_component_name(self):
        # type: () -> Optional[str]
        """
        Retrieves the name of the component associated to the event

        :return: the name of the component
        """
        return self.__component_name

    def get_factory_name(self):
        # type: () -> str
        """
        Retrieves the name of the factory associated to the event

        :return: the name of the component factory
        """
        return self.__factory_name

    def get_kind(self):
        # type: () -> int
        """
        Retrieves the kind of event

        :return: the kind of event
        """
        return self.__kind



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Utility methods for SSL

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import ssl

try:
    from ssl import _RESTRICTED_SERVER_CIPHERS
except ImportError:
    # Restricted and more secure ciphers for the server side, from Python 3.5
    _RESTRICTED_SERVER_CIPHERS = (
        "ECDH+AESGCM:DH+AESGCM:ECDH+AES256:DH+AES256:ECDH+AES128:DH+AES:"
        "ECDH+HIGH:DH+HIGH:ECDH+3DES:DH+3DES:RSA+AESGCM:RSA+AES:RSA+HIGH:"
        "RSA+3DES:!aNULL:!eNULL:!MD5:!DSS:!RC4"
    )

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


def wrap_socket(socket, certfile, keyfile, password=None):
    """
    Wraps an existing TCP socket and returns an SSLSocket object

    :param socket: The socket to wrap
    :param certfile: The server certificate file
    :param keyfile: The server private key file
    :param password: Password for the private key file (Python >= 3.3)
    :return: The wrapped socket
    :raise SSLError: Error wrapping the socket / loading the certificate
    :raise OSError: A password has been given, but ciphered key files are not
                    supported by the current version of Python
    """
    # Log warnings when some
    logger = logging.getLogger("ssl_wrap")

    def _password_support_error():
        """
        Logs a warning and raises an OSError if a password has been given but
        Python doesn't support ciphered key files.

        :raise OSError: If a password has been given
        """
        if password:
            logger.error(
                "The ssl.wrap_socket() fallback method doesn't "
                "support key files with a password."
            )
            raise OSError(
                "Can't decode the SSL key file: "
                "this version of Python doesn't support it"
            )

    try:
        # Prefer the default context factory, as it will be updated to reflect
        # security issues (Python >= 2.7.9 and >= 3.4)
        default_context = ssl.create_default_context()
    except AttributeError:
        default_context = None

    try:
        # Try to equivalent to create_default_context() in Python 3.5
        # Create an SSL context and set its options
        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)

        if default_context is not None:
            # Copy options
            context.options = default_context.options
        else:
            # Set up the context as create_default_context() does in Python 3.5
            # SSLv2 considered harmful
            # SSLv3 has problematic security
            context.options |= ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3

        # disallow ciphers with known vulnerabilities
        context.set_ciphers(_RESTRICTED_SERVER_CIPHERS)

        try:
            # Load the certificate, with a password
            context.load_cert_chain(certfile, keyfile, password)
        except TypeError:
            # The "password" argument isn't supported
            # Check support for key file password
            _password_support_error()

            # Load the certificate, without the password argument
            context.load_cert_chain(certfile, keyfile)

        # Return the wrapped socket
        return context.wrap_socket(socket, server_side=True)

    except AttributeError as ex:
        # Log a warning to advise the user of possible security holes
        logger.warning(
            "Can't create a custom SSLContext. "
            "The server should be considered insecure."
        )
        logger.debug("Missing attribute: %s", ex)

    # Check support for key file password
    _password_support_error()

    # Fall back to the "old" wrap_socket method
    return ssl.wrap_socket(
        socket, server_side=True, certfile=certfile, keyfile=keyfile
    )



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
MQTT client utility: Tries to hide Paho client details to ease MQTT usage.
Reconnects to the MQTT server automatically.

This module depends on the paho-mqtt package (ex-mosquitto), provided by the
Eclipse Foundation: see http://www.eclipse.org/paho

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import os
import sys
import threading

# MQTT client
import paho.mqtt.client as paho

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


class MqttClient(object):
    """
    Remote Service discovery provider based on MQTT
    """

    def __init__(self, client_id=None):
        """
        Sets up members

        :param client_id: ID of the MQTT client
        """
        # No ID
        if not client_id:
            # Randomize client ID
            self._client_id = self.generate_id()
        elif len(client_id) > 23:
            # ID too large
            _logger.warning(
                "MQTT Client ID '%s' is too long (23 chars max): "
                "generating a random one",
                client_id,
            )
            # Keep the client ID as it might be accepted
            self._client_id = client_id
        else:
            # Keep the ID as is
            self._client_id = client_id

        # Reconnection timer
        self.__timer = threading.Timer(5, self.__reconnect)

        # Publication events
        self.__in_flight = {}

        # MQTT client
        self.__mqtt = paho.Client(self._client_id)

        # Give access to Paho methods to configure TLS
        self.tls_set = self.__mqtt.tls_set

        # Paho callbacks
        self.__mqtt.on_connect = self.__on_connect
        self.__mqtt.on_disconnect = self.__on_disconnect
        self.__mqtt.on_message = self.__on_message
        self.__mqtt.on_publish = self.__on_publish

    @property
    def raw_client(self):
        """
        Returns the raw client object, depending on the underlying library
        """
        return self.__mqtt

    @staticmethod
    def on_connect(client, result_code):
        """
        User callback: called when the client is connected

        :param client: The Pelix MQTT client which connected
        :param result_code: The MQTT result code
        """
        pass

    @staticmethod
    def on_disconnect(client, result_code):
        """
        User callback: called when the client is disconnected

        :param client: The Pelix MQTT client which disconnected
        :param result_code: The MQTT result code
        """
        pass

    @staticmethod
    def on_message(client, message):
        """
        User callback: called when the client has received a message

        :param client: The Pelix MQTT client which received a message
        :param message: The MQTT message
        """
        pass

    @classmethod
    def generate_id(cls, prefix="pelix-"):
        """
        Generates a random MQTT client ID

        :param prefix: Client ID prefix (truncated to 8 chars)
        :return: A client ID of 22 or 23 characters
        """
        if not prefix:
            # Normalize string
            prefix = ""
        else:
            # Truncate long prefixes
            prefix = prefix[:8]

        # Prepare the missing part
        nb_bytes = (23 - len(prefix)) // 2

        random_bytes = os.urandom(nb_bytes)
        if sys.version_info[0] >= 3:
            random_ints = [char for char in random_bytes]
        else:
            random_ints = [ord(char) for char in random_bytes]

        random_id = "".join("{0:02x}".format(value) for value in random_ints)
        return "{0}{1}".format(prefix, random_id)

    @classmethod
    def topic_matches(cls, subscription_filter, topic):
        """
        Checks if the given topic matches the given subscription filter

        :param subscription_filter: A MQTT subscription filter
        :param topic: A topic
        :return: True if the topic matches the filter
        """
        return paho.topic_matches_sub(subscription_filter, topic)

    @property
    def client_id(self):
        """
        The MQTT client ID
        """
        return self._client_id

    def set_credentials(self, username, password):
        """
        Sets the user name and password to be authenticated on the server

        :param username: Client username
        :param password: Client password
        """
        self.__mqtt.username_pw_set(username, password)

    def set_will(self, topic, payload, qos=0, retain=False):
        """
        Sets up the will message

        :param topic: Topic of the will message
        :param payload: Content of the message
        :param qos: Quality of Service
        :param retain: The message will be retained
        :raise ValueError: Invalid topic
        :raise TypeError: Invalid payload
        """
        self.__mqtt.will_set(topic, payload, qos, retain=retain)

    def connect(self, host="localhost", port=1883, keepalive=60):
        """
        Connects to the MQTT server. The client will automatically try to
        reconnect to this server when the connection is lost.

        :param host: MQTT server host
        :param port: MQTT server port
        :param keepalive: Maximum period in seconds between communications with
                          the broker
        :raise ValueError: Invalid host or port
        """
        # Disconnect first (it also stops the timer)
        self.disconnect()

        # Prepare the connection
        self.__mqtt.connect(host, port, keepalive)

        # Start the MQTT loop
        self.__mqtt.loop_start()

    def disconnect(self):
        """
        Disconnects from the MQTT server
        """
        # Stop the timer
        self.__stop_timer()

        # Unlock all publishers
        for event in self.__in_flight.values():
            event.set()

        # Disconnect from the server
        self.__mqtt.disconnect()

        # Stop the MQTT loop thread
        # Use a thread to avoid a dead lock in Paho
        thread = threading.Thread(target=self.__mqtt.loop_stop)
        thread.daemon = True
        thread.start()

        # Give it some time
        thread.join(4)

    def publish(self, topic, payload, qos=0, retain=False, wait=False):
        """
        Sends a message through the MQTT connection

        :param topic: Message topic
        :param payload: Message content
        :param qos: Quality of Service
        :param retain: Retain flag
        :param wait: If True, prepares an event to wait for the message to be
                     published
        :return: The local message ID, None on error
        """
        result = self.__mqtt.publish(topic, payload, qos, retain)
        if wait and not result[0]:
            # Publish packet sent, wait for it to return
            self.__in_flight[result[1]] = threading.Event()
            _logger.debug("Waiting for publication of %s", topic)

        return result[1]

    def wait_publication(self, mid, timeout=None):
        """
        Wait for a publication to be validated

        :param mid: Local message ID (result of publish)
        :param timeout: Wait timeout (in seconds)
        :return: True if the message was published, False if timeout was raised
        :raise KeyError: Unknown waiting local message ID
        """
        return self.__in_flight[mid].wait(timeout)

    def subscribe(self, topic, qos=0):
        """
        Subscribes to a topic on the server

        :param topic: Topic filter string(s)
        :param qos: Desired quality of service
        :raise ValueError: Invalid topic or QoS
        """
        self.__mqtt.subscribe(topic, qos)

    def unsubscribe(self, topic):
        """
        Unscribes from a topic on the server

        :param topic: Topic(s) to unsubscribe from
        :raise ValueError: Invalid topic parameter
        """
        self.__mqtt.unsubscribe(topic)

    def __start_timer(self, delay):
        """
        Starts the reconnection timer

        :param delay: Delay (in seconds) before calling the reconnection method
        """
        self.__timer = threading.Timer(delay, self.__reconnect)
        self.__timer.daemon = True
        self.__timer.start()

    def __stop_timer(self):
        """
        Stops the reconnection timer, if any
        """
        if self.__timer is not None:
            self.__timer.cancel()
            self.__timer = None

    def __reconnect(self):
        """
        Tries to connect to the MQTT server
        """
        # Cancel the timer, if any
        self.__stop_timer()

        try:
            # Try to reconnect the server
            result_code = self.__mqtt.reconnect()
            if result_code:
                # Something wrong happened
                message = "Error connecting the MQTT server: {0} ({1})".format(
                    result_code, paho.error_string(result_code)
                )
                _logger.error(message)
                raise ValueError(message)
        except Exception as ex:
            # Something went wrong: log it
            _logger.error("Exception connecting server: %s", ex)
        finally:
            # Prepare a reconnection timer. It will be cancelled by the
            # on_connect callback
            self.__start_timer(10)

    def __on_connect(self, client, userdata, flags, result_code):
        # pylint: disable=W0613
        """
        Client connected to the server

        :param client: Connected Paho client
        :param userdata: User data (unused)
        :param flags: Response flags sent by the broker
        :param result_code: Connection result code (0: success, others: error)
        """
        if result_code:
            # result_code != 0: something wrong happened
            _logger.error(
                "Error connecting the MQTT server: %s (%d)",
                paho.connack_string(result_code),
                result_code,
            )
        else:
            # Connection is OK: stop the reconnection timer
            self.__stop_timer()

        # Notify the caller, if any
        if self.on_connect is not None:
            try:
                self.on_connect(self, result_code)
            except Exception as ex:
                _logger.exception("Error notifying MQTT listener: %s", ex)

    def __on_disconnect(self, client, userdata, result_code):
        # pylint: disable=W0613
        """
        Client has been disconnected from the server

        :param client: Client that received the message
        :param userdata: User data (unused)
        :param result_code: Disconnection reason (0: expected, 1: error)
        """
        if result_code:
            # rc != 0: unexpected disconnection
            _logger.error(
                "Unexpected disconnection from the MQTT server: %s (%d)",
                paho.connack_string(result_code),
                result_code,
            )

            # Try to reconnect
            self.__stop_timer()
            self.__start_timer(2)

        # Notify the caller, if any
        if self.on_disconnect is not None:
            try:
                self.on_disconnect(self, result_code)
            except Exception as ex:
                _logger.exception("Error notifying MQTT listener: %s", ex)

    def __on_message(self, client, userdata, msg):
        # pylint: disable=W0613
        """
        A message has been received from a server

        :param client: Client that received the message
        :param userdata: User data (unused)
        :param msg: A MQTTMessage bean
        """
        # Notify the caller, if any
        if self.on_message is not None:
            try:
                self.on_message(self, msg)
            except Exception as ex:
                _logger.exception("Error notifying MQTT listener: %s", ex)

    def __on_publish(self, client, userdata, mid):
        # pylint: disable=W0613
        """
        A message has been published by a server

        :param client: Client that received the message
        :param userdata: User data (unused)
        :param mid: Message ID
        """
        try:
            self.__in_flight[mid].set()
        except KeyError:
            pass
